/***************************************************************************
 *
 * This is a part of AntMon project
 *
 *            ant_agent.cpp
 *
 *  Tue Aug 22 14:33:28 2006
 *  Copyright  2006  Sergey Zhumatiy
 *  serg@parallel.ru
 ****************************************************************************/
/*
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include <unistd.h>
#include <signal.h>
#include <sys/wait.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/socket.h>
#include <sys/resource.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <fcntl.h>
#include <resolv.h>
#include <netdb.h>
#include <time.h>
#include <errno.h>
#include <stdlib.h>
#include <string.h>

#include <string>
#include <list>
#include <vector>
#include <queue>

#include "ant_mod.hpp"
#include "ant_sens.hpp"
#include "ant_conn.hpp"
#include "ant_log.hpp"
#include "ant_time.hpp"
#include "ant_net.hpp"
#include "ant_exe.hpp"

#ifndef O_LARGEFILE
#define O_LARGEFILE 0
#endif

#define ANT_MOD_RETRY_INTERVAL 600
#define MAX_SERV 16

using namespace std;


void sig_log_handler(int);
void sig_urg_handler(int);
void exe_sig_chld_handler(int);
void sig_exit_handler(int);

//void load_config();
void create_connections();
int  GetNonblockSocket();
int  GetAddress(struct sockaddr_in *addr, char *host, int port);
void process_message(int i);
void process_timed_out(Ant_agent_sensor *sens);
void process_mod_answer(Ant_agent_module *mod);
void daemonize();
void gentle_exit(int code);
#define CHLD_INTERVAL 5

void on_module_fail(Ant_agent_module *mod);
void on_server_fail(int i);

/*
int build_server_socket(struct addrinfo *ai);
int multi_accept(int *server_s, int nb_server_s,
                 struct sockaddr *addr=NULL, socklen_t *len=NULL);
*/

long int new_id();

/////////////////////////////////////////////////////////////////
//
//  Global variabbles
//
/////////////////////////////////////////////////////////////////

//#define DEBUG(x) x
#define DEBUG(x) ;

#define ANT_PORT 8898

#define MAX_ID 2147483647
//             = 2**31-1

#define MAX_BUF 10240

#define MAX_CLIENTS 10

//typedef vector<class Ant_param *> Ant_param_vector;

// list of all connections
vector<class Ant_connection *>          Connections;
// list of sent requests
//vector<list<class Ant_agent_sensor *> *>  Waited_ids_by_ip;

list<class Ant_agent_module *>                Modules;


struct waiter{
  int        conn_n;
  long int   id;
  Ant_agent_sensor *sens;
};

// list of sent requests
list<struct waiter *>                   Waiters;

// copy ot time()
time_t         sensors_timeout;
// sleep interval
struct timeval sleep_time;

static long int _id=1;
static char tmpbuf[MAX_BUF];

//static int sd; // socket

short int listen_port=ANT_PORT;
int fresh_time=3;

/////////////////////////////////////////////////////////////////
//
//  Functions
//
/////////////////////////////////////////////////////////////////


char def_log[]="Ant agent";

//  Main
//
/////////////////////////////////////////////////////////////////
int main(int argc, char *argv[]){
  int num_of_conn=0;
  int i,j; //,max,flag;
  //int interval_add;
  int daemon=0;

  FILE *mod_errors=NULL;
  const char *err;
  //char port_str[NI_MAXSERV];
  //int  code;
  int  server_s[MAX_SERV];
  int  nb_server_s = 0;
  time_t last_take_chld=0;

  int client_sd;
  int syslog_flag;

  struct sockaddr addr;
  socklen_t addr_size; // = sizeof(addr);

  Ant_agent_sensor *sens;
  list<class Ant_agent_module *>::iterator mod_end;

  syslog_flag   = 0;
  log_to_syslog = 0;
  log_name=def_log;

  // arguments parsing
  for(i=1;i<argc;++i){

    // help?
    if(!strcmp(argv[i],"-h")){
      printf("Usage: ant_agent [options]\n");
      printf("       -h          help\n");
      printf("       -d          daemonize\n");
      printf("       -s          log to syslog\n");
      printf("       -f          log to stderr\n");
      printf("       -n prefix   prefix for syslog strings\n");
      printf("       -g level    loglevel (see README file)\n");
      printf("       -e file     dump all modules errors to file\n");
      printf("       -l file     log to file\n\n");
      gentle_exit(0);
    }

    // daemonize?
    if(!strcmp(argv[i],"-d")){
      daemon=1;
      if(syslog_flag==0){
          log_to_syslog=1;
      }
      continue;
    }

    // log to syslog?
    if(!strcmp(argv[i],"-s")){
      log_to_syslog=1;
      continue;
    }

    // log to stderr?
    if(!strcmp(argv[i],"-f")){
      log_to_syslog=0;
      syslog_flag=1;
      continue;
    }

    // syslog name
    if(!strcmp(argv[i],"-n")){
      if(++i<argc){
        if(log_name)
          delete log_name;
        log_name=new char[strlen(argv[i])+1];
        strcpy(log_name,argv[i]);
      }
      else{
          LOG(ALOG_WARN,("-n option requires argument."));
      }
      continue;
    }

    // log level
    if(!strcmp(argv[i],"-g")){
      if(++i<argc){
        sscanf(argv[i],"%d",&log_level);
      }
      else{
          LOG(ALOG_WARN,("-g option requires argument."));
      }
      continue;
    }

    // log to file
    if(!strcmp(argv[i],"-l")){
      if(++i<argc){
        log_to_syslog=0;
        syslog_flag=1;
        j=open(argv[i],O_WRONLY|O_CREAT|O_APPEND|O_LARGEFILE,S_IRWXU|S_IRGRP);
        if(j>0){
          close(2);
          dup2(j,2);
          close(j);
        }
        else
          LOG(ALOG_WARN,("Cannot open %s: %s",argv[i], strerror(errno)));
      }
      else{
          LOG(ALOG_WARN,("-l option requires argument."));
      }
      continue;
    }

    // modules errors to file
    if(!strcmp(argv[i],"-e")){
      if(++i<argc){
        mod_errors=fopen(argv[i],"a+");
        if(mod_errors!=NULL){
            chmod(argv[i],S_IRUSR|S_IWUSR|S_IRGRP);
        }
        else
          LOG(ALOG_WARN,("Cannot open %s: %s",argv[i], strerror(errno)));
      }
      else{
          LOG(ALOG_WARN,("-e option requires argument."));
      }
      continue;
    }
  }

  if(mod_errors==NULL){
      mod_errors=fopen("/dev/null","a+");
      if(mod_errors==NULL){
        LOG(ALOG_ALL,("Cannot open /dev/null! Exit."));
         gentle_exit(20);
      }
  }

  LOG(ALOG_ALL,("Starting..."));

  if(daemon)
    daemonize();

/*  struct sigaction act_chld;
  struct sigaction act_urg;
*/
//  log_to_syslog=0;

  signal(SIGQUIT,sig_exit_handler);
  signal(SIGTERM,sig_exit_handler);
  signal(SIGABRT,sig_exit_handler);
  signal(SIGSEGV,sig_exit_handler);
  signal(SIGXCPU,sig_exit_handler);
  signal(SIGFPE,sig_exit_handler);
  signal(SIGBUS,sig_exit_handler);

  signal(SIGPIPE,sig_log_handler);
  signal(SIGURG, sig_urg_handler);
  signal(SIGCHLD,exe_sig_chld_handler);


  Ant_connection::send_retr=3;
  Ant_connection::sleep_time.tv_sec=0;
  Ant_connection::sleep_time.tv_usec=500000;
  Ant_agent_module::send_retr=3;
  Ant_agent_module::sleep_time.tv_sec=0;
  Ant_agent_module::sleep_time.tv_usec=500000;


    /*listen_port=ANT_PORT;*/
    nb_server_s=build_server_socket(ANT_PORT, server_s, MAX_SERV);
    if (nb_server_s < 1) {
        LOG(ALOG_WARN | ALOG_NET, ("no suitable listen address found\n"));
        gentle_exit(2);
    }

  sensors_timeout=10;
  /*
    f=fopen("/tmp/monitor.pid","w");
    //    f=fopen("/var/run/monitor.pid","w");
    fprintf(f,"%d",getpid());
    printf("My PID is %d\n",getpid());
    fclose(f);

  */

  //  load_config();

/*

  // make listening socket
  if ( (sd = socket(PF_INET, SOCK_STREAM, 0)) < 0 ){
    LOG(ALOG_WARN,("Cannot create listening socket"));
    gentle_exit(1);
  }
  flag=1;
  setsockopt(sd, SOL_SOCKET, SO_REUSEADDR, &flag, sizeof(int));
  bzero(&addr, sizeof(addr));
  addr.sin_family = AF_INET;
  addr.sin_port = htons(listen_port);
  addr.sin_addr.s_addr = htonl(INADDR_ANY);
  if ( bind(sd, (struct sockaddr*)&addr, sizeof(addr)) != 0 ){
    LOG(ALOG_WARN,("Cannot bind to port %d (%s)", listen_port, strerror(errno)));
    gentle_exit(2);
  }
  if ( listen(sd, MAX_CLIENTS) != 0 ){
    LOG(ALOG_WARN,("Cannot listen port %d (%s)", listen_port, strerror(errno)));
    gentle_exit(3);
  }
  if (fcntl(sd,F_SETFL,O_NONBLOCK|fcntl(sd,F_GETFL,O_NONBLOCK))!=0){
    LOG(ALOG_WARN,("Cannot make port %d nonblocked (%s)", listen_port,
        strerror(errno)));
    gentle_exit(4);
  }
*/

//  LOG(ALOG_INFO,("Listen on %d",listen_port));

  //
  //
  // ===================== MAIN LOOP =======================================
  //
  //

  for(;;) {
    //get current time
    fix_time();

    // call waitpid for childs sometimes...
    if((global_time-last_take_chld>CHLD_INTERVAL) ||
        kill_flag!=0){
        last_take_chld=global_time;
        take_childs();
    }

    //
    // try to accept new connections
    //
    //!!!struct sockaddr addr;

    client_sd = multi_accept(server_s, nb_server_s, &addr, &addr_size);

    if (client_sd >=0) {
        // accepted...

        char sport[NI_MAXSERV];
        char serv_name[NI_MAXHOST];

        /*serv_name = inet_ntoa(addr.sin_addr); */
        //get_name_and_port(mon_sd, serv_name, sport);
        int code;
        if ((code = getnameinfo(&addr, addr_size, serv_name,
                                NI_MAXHOST, sport, NI_MAXSERV,
                                NI_NUMERICHOST|NI_NUMERICSERV))){
            LOG(ALOG_WARN | ALOG_NET, ("Error on getaddrinfo: %s",
                                       gai_strerror(code)));
            if (code == EAI_SYSTEM) {
                LOG(ALOG_WARN | ALOG_NET,
                ("Details: %s", strerror(errno)));
            }
            strcpy(serv_name,"undefined_host");
        }

        LOG(ALOG_NET,("Connected %s:%s at %d",serv_name, sport, client_sd));

        // add to connections pool
        Ant_connection *conn=new Ant_connection;
        Connections.push_back(conn);

        // add corresponding list of parameters to await
        //      list<class Ant_agent_sensor *> *lst= new list<class Ant_agent_sensor *>;
        //      Waited_ids_by_ip.push_back(lst);

        // set apropriate parameters for conenction...
        conn->init(serv_name);
        conn->fd=client_sd;
        if (fcntl(client_sd,F_SETFL,O_NONBLOCK|
            fcntl(client_sd,F_GETFL,O_NONBLOCK))!=0){
          LOG(ALOG_WARN,("Cannot make port %d nonblocked (%s)",
              client_sd, strerror(errno)));
          gentle_exit(5);
        }
        conn->connected=1;

    }

    //
    // Process server connections
    //
    num_of_conn=Connections.size();
    for(i=0;i<num_of_conn;++i) {
      Ant_connection *conn=Connections[i];

      // read messages
      while(conn->getMessage()){
        process_message(i);
        conn->clean();
      }

      // test associated parameters for timeouts
      list<waiter *>::iterator end=Waiters.end();
      for(list<waiter *>::iterator li=Waiters.begin();li!=end;){
        sens=(*li)->sens;

        // timed out?
        if(sens->timeout<global_time){

          // send answer to server
          process_timed_out(sens);

          //!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
          //!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
          //!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

          // delete this parameter from list of awaited
          list<waiter *>::iterator del=li++;
          Waiters.erase(del);
        }
        else
          ++li;
      }

      // flush data back to server
      //if(conn->out_pos)
      //  log("flush");
      if(conn->flush()){

        // error with server connection
        on_server_fail(i);
      }

    }//for all connections

    //
    //  Process modules requests and answers
    //
    mod_end=Modules.end();
    for(list<class Ant_agent_module *>::iterator li=Modules.begin();
        li!=mod_end;++li){
      Ant_agent_module *mod=*li;

      // is module active?
      if(!mod->active)
        continue;

      if(mod->out_pos)
        LOG(ALOG_DEBUG,("mod_flush (%s)",mod->name));
      if(mod->flush()){

        // error on module interaction
        on_module_fail(mod);
        continue;
      }

      // is there an answer?
      while(mod->readMessage()){
        LOG(ALOG_DEBUG,("mod answered (%s): %s",mod->name,mod->buffer));
        process_mod_answer(mod);
        mod->clean();
      }
      if(mod->pos)
        LOG(ALOG_DEBUG,("readed(%s) %d",mod->name,mod->pos));

      // is there an error?
      if(mod->error){
        LOG(ALOG_WARN,("Error on module %s: %s", mod->name,
            strerror(mod->error)));
        mod->finalize();
        mod->next_try=global_time+ANT_MOD_RETRY_INTERVAL;
      }

      if((err=mod->readErrors())!=NULL){
          fprintf(mod_errors,"[%s] %s\n",mod->name, err);
          fflush(mod_errors);
      }

    }

    //    DEBUG(LOG(,((long int)*Ant_agent_module::global_ids["test_mod.test13"]);))
    // sleep
    sleep_time.tv_sec=0;
    sleep_time.tv_usec=500*1000;
    select(0,NULL,NULL,NULL,&sleep_time);

    //    log("mark");

  }// main loop

  fclose(mod_errors);
  return 1;
}


void sig_log_handler(int sig){
  LOG(ALOG_WARN,("Signal %d received. Ignore.",sig));
}

void sig_urg_handler(int sig){
  LOG(ALOG_WARN,("Signal %d received. Ignore.",sig));
}

void sig_exit_handler(int sig){
  LOG(ALOG_WARN,("Signal %d received. Exit now.",sig));

  // kill modules
  kill(0,SIGQUIT);
  gentle_exit(0);
}

int GetNonblockSocket(){
  int sock;
  if( (sock=socket(PF_INET,SOCK_STREAM,0)) <0 ) {
    LOG(ALOG_WARN,("Error creating socket: %s",strerror(errno)));
    return(-1);
  }
  if (fcntl(sock,F_SETFL,O_NONBLOCK)!=0) {
    LOG(ALOG_WARN,("Error creating socket: %s",strerror(errno)));
    return(-1);
  }
  return sock;
}

/*********************************************************************/
/*** GetAddress                                                    ***/
/***                                                               ***/
/*** Use the DNS to resolve host name.  (If "host" is an IP        ***/
/*** address, no resolution is done, and the address is converted  ***/
/*** for you.)  Fills the sockaddr_in struct.                      ***/
/*** Returns zero if successful; negative value, otherwise.        ***/
/*********************************************************************/
int GetAddress(struct sockaddr_in *addr, char *host, int port){
  //  struct hostent *ent;
  struct addrinfo hint,*res;
  struct sockaddr_in *in;
  int code;
/*
       struct addrinfo {
           int     ai_flags;
           int     ai_family;
           int     ai_socktype;
           int     ai_protocol;
           size_t  ai_addrlen;
           struct sockaddr *ai_addr;
           char   *ai_canonname;
           struct addrinfo *ai_next;
       };

*/
  hint.ai_family=PF_INET;
  hint.ai_socktype=0;
  hint.ai_protocol=0;

  bzero(addr, sizeof(*addr));
  addr->sin_family = AF_INET;
  addr->sin_port = htons(port);
  if ( host != 0 ) {
    if((code=getaddrinfo(host, NULL,&hint,&res))!=0){
      LOG(ALOG_WARN,("Cannot resolve %s (%s)",host,gai_strerror(code)));
      return -1;
    }

    in=(struct sockaddr_in *)&(res->ai_addr);
    addr->sin_addr.s_addr=in->sin_addr.s_addr;
    freeaddrinfo(res);
    return 0;
  }
  addr->sin_addr.s_addr = htonl(INADDR_ANY);
  return 0;
}

// generate new almost unique id
long int new_id(){
  if(++_id>MAX_ID){
    _id=1;
    return 1;
  }
  return _id;
}


//
//  Process incoming message from servr
//
void process_message(int conn_n){
  long int id, sens_id;
  Ant_int  *ip;
  int i, j; //, fail_dep, bad_val;
  Ant_agent_sensor *sens;
  Ant_connection *conn=Connections[conn_n];

  LOG(ALOG_DEBUG,("Got from server: '%s'",conn->buffer));

  // read id
  id=0;
  for(i=0;conn->buffer[i]>='0' && conn->buffer[i]<='9'; ++i){
    id=id*10+conn->buffer[i]-'0';
  }

  // check format (must be ':' delimiter)
  if(conn->buffer[i]!=':' || conn->buffer[i+2]!=':'){
    LOG(ALOG_WARN,("Bad request from %s (%s)",conn->name.c_str(),conn->buffer));
    return;
  }
  ++i;

  // read action
  switch(conn->buffer[i]){
    // 'add' request (register sensor)
  case 'a':
    // read parameter name
    i+=2;
    for(j=i;conn->buffer[j]!='\n';)
      ++j;

    conn->buffer[j]='\0';


    ip=Ant_agent_module::global_ids[conn->buffer+i];

    // not registered?
    if(ip==NULL){
      char *point=conn->buffer+i;
      char *space=point;

      Ant_agent_module::global_ids.erase(conn->buffer+i);

      // try to register it

      // cut module name and args (search LAST ' ')
      while(*point){
    // replace slashes only in module name
        if((*point=='/') && (space==conn->buffer+i))
          *point='_';
        else if(*point==' ')
          space=point;
        ++point;
      }

      // bad sensor name?
      if(space==point){
        LOG(ALOG_SENS,("No such sensor: %s. Bad sensor name",conn->buffer+i));
        snprintf(tmpbuf,MAX_BUF-1,"%ld:a:x\n\n",id);
        tmpbuf[MAX_BUF-1]='\0';
        conn->sendMessage(tmpbuf);
        break;
      }

      // check if this module already exists
      *space='\0';

      list<class Ant_agent_module *>::iterator end=Modules.end();
      list<class Ant_agent_module *>::iterator it;
      for(it=Modules.begin();it!=end;++it){

        // module already loaded?
        if(!strcmp((*it)->name,conn->buffer+i)){

          // is module universal?
          string str=conn->buffer+i; //copy module name (without sensor!)
          str+=" *";
          ip=Ant_agent_module::global_ids[str];

          // restore full name (with sensor)
          *space=' ';
          if(ip!=NULL){

            // yes, it's universal!
            //            (*it)->register_sensor(conn->buffer+i);
            (*it)->register_sensor(space+1);
            ip=Ant_agent_module::global_ids[conn->buffer+i];
          }
          else{
            // module is loaded, but do not provides this...
            Ant_agent_module::global_ids.erase(conn->buffer+i);
            LOG(ALOG_SENS,("No such sensor: %s. Module do not provide it.",conn->buffer+i));
            snprintf(tmpbuf,MAX_BUF-1,"%ld:a:x\n\n",id);
            tmpbuf[MAX_BUF-1]='\0';
            conn->sendMessage(tmpbuf);
            break;
          }
        }
      }// check if module already loaded


      // found sensor?
      if(ip==NULL){

        // sensor not found and module does not exists yet
        // register new module!
        Ant_agent_module *mod =new Ant_agent_module(conn->buffer+i);

        mod->data_sent=0;

        // parse and create arguments for module
        {
          //char **params;
          //int  pi,param_num=0;
          point=conn->buffer+i;

          // init module!
          mod->init(conn->buffer+i);

        }
        // restore sensor name
        *space=' ';

        // module fails?
        if(!mod->active){
          Ant_agent_module::global_ids.erase(conn->buffer+i);
          LOG(ALOG_SENS,("No such sensor: %s. Module fails.",conn->buffer+i));
          snprintf(tmpbuf,MAX_BUF-1,"%ld:a:x\n\n",id);
          tmpbuf[MAX_BUF-1]='\0';
          conn->sendMessage(tmpbuf);
          delete mod;
          break;
        }

        Modules.push_back(mod);

        // check, if register is successfull
        ip=Ant_agent_module::global_ids[conn->buffer+i];

        // direct register failed?
        if(ip==NULL){
          Ant_agent_module::global_ids.erase(conn->buffer+i);

          // is module univarsal?
          //
          *space='\0';
          string str=conn->buffer+i; //copy module name (without sensor!)
          str+=" *";
          ip=Ant_agent_module::global_ids[str];

          // restore full name (with sensor)
          *space=' ';
          if(ip!=NULL){

            // yes, it's universal!
            //            mod->register_sensor(conn->buffer+i);
            mod->register_sensor(space+1);
            ip=Ant_agent_module::global_ids[conn->buffer+i];
          }
          else{
            // module is loaded, but do not provides this...
            Ant_agent_module::global_ids.erase(conn->buffer+i);
            LOG(ALOG_SENS,("No such sensor: %s",conn->buffer+i));
            snprintf(tmpbuf,MAX_BUF-1,"%ld:a:x\n\n",id);
            tmpbuf[MAX_BUF-1]='\0';
            conn->sendMessage(tmpbuf);
            break;
          }
        } // direct register check
      } //
    } // registering new module

    sens_id=(long int)*ip;
    LOG(ALOG_INFO,("Sensor: '%s' = %ld",conn->buffer+i,sens_id));

    // registered
    snprintf(tmpbuf,MAX_BUF-1,"%ld:a:%ld\n\n",id,sens_id);
    tmpbuf[MAX_BUF-1]='\0';
    conn->sendMessage(tmpbuf);

    break; // 'add' request

    // 'get' request
  case 'g':

    // read parameter id
    i+=2;
    sens_id=0;
    for(;conn->buffer[i]>='0' && conn->buffer[i]<='9'; ++i){
      sens_id=sens_id*10+conn->buffer[i]-'0';
    }

    // get pionter to ag_id
    sens=(Ant_agent_sensor *)Ant_agent_sensor::all[sens_id];

    // is it valid?
    if(sens==NULL){
      Ant_agent_sensor::all.erase(sens_id);
      LOG(ALOG_WARN,("Id does not exists: '%s'",conn->buffer));
      break;
    }
    // second check
    if(sens->id){

      // check its value (may be it's fresh?)
      if(sens->last_time+fresh_time>=global_time){

        // yes. Just send value
        snprintf(tmpbuf,MAX_BUF-1,"%ld:g:%f\n\n",id,sens->value);
        tmpbuf[MAX_BUF-1]='\0';
        conn->sendMessage(tmpbuf);
        break;
      }

      // check if modules is alive
      if(!sens->module->active){
        LOG(ALOG_WARN,("Module is disabled"));

        // answer
        snprintf(tmpbuf,MAX_BUF-1,"%ld:g:x\n\n",id);
        tmpbuf[MAX_BUF-1]='\0';
        conn->sendMessage(tmpbuf);

        // try to restart module
        if(global_time>sens->module->next_try){
          LOG(ALOG_WARN,("Try to restart it!"));
          string tempstr=sens->module->exe;
          sens->module->init(tempstr.c_str());
          sens->module->next_try=global_time+ANT_MOD_RETRY_INTERVAL;
        }
        break;

      }

      LOG(ALOG_DEBUG,("request to module %s",sens->module->name));

      // request the value
      if(sens->module->request(sens))
        break; // not sent to module

      sens->timeout=global_time+sensors_timeout;
      // add waiter record
      struct waiter *w=new waiter;
      w->sens=sens;
      w->conn_n=conn_n;
      w->id=id;
      Waiters.push_back(w);
      LOG(ALOG_DEBUG,("Added waiter: id=%ld, conn=%d (total %d)",id,conn_n,Waiters.size()));

    }
    // not valid
    else{

      Ant_agent_sensor::all.erase(sens_id);
      LOG(ALOG_WARN,("Bad id in request '%s'",conn->buffer));
    }

    break;  // 'g' request

  case 'p': //ping

    LOG(ALOG_DEBUG,("ping received"));

    snprintf(tmpbuf,MAX_BUF-1,"%ld:p:\n\n",id);
    tmpbuf[MAX_BUF-1]='\0';
    conn->sendMessage(tmpbuf);
    break;

  default:
    LOG(ALOG_WARN,("Unknown request: %s",conn->buffer));
  }
}

void process_timed_out(Ant_agent_sensor *sens){
  LOG(ALOG_WARN,("Sensor timed out: %s->%s",sens->module->name,sens->name));
}

// answer from module received
void process_mod_answer(Ant_agent_module *mod){
  int i,start;
  Ant_agent_sensor *sens;

  //current line offset
  start=0;

  // set up iterator
  list<Ant_agent_sensor *>::iterator end=mod->sent_ids.end();
  list<Ant_agent_sensor *>::iterator li=mod->sent_ids.begin();

  //get all lines from answer
  for(i=0;mod->buffer[i];++i){

    // line end?
    if(mod->buffer[i]=='\n'){

      LOG(ALOG_DEBUG,("Process mod line:'%s'",mod->buffer+start));
      // parse answer
      if(li==end){
        LOG(ALOG_WARN,
            ("Error: module %s returned more values than i ask it. rest of line:%s.",mod->name,mod->buffer+start)
            );
        break;
      }

      sens=*li;

      //DEBUG(LOG(,((void *)sens);))

      ++li;
      if(sens==NULL){
        LOG(ALOG_WARN,("Error: module %s returned unexpected value?",mod->name));
        break;
      }

      // test answer for fail
      if(strcmp(mod->buffer+start,"nan")==0){
        LOG(ALOG_EXT,("Module %s returned NaN.",mod->name));
        sens->is_nan=1;
      }
      else{
        sens->is_nan=0;

        // get value
        if(sscanf(mod->buffer+start,"%f",&(sens->value))<1){
          LOG(ALOG_EXT,("Module %s returned bad value (%s).",mod->name,mod->buffer));
        }
      }

      // send it back to server(s)
      // cycle by waiters
      LOG(ALOG_DEBUG,("total waiters: %d",Waiters.size()));
      list<struct waiter *>::iterator wend=Waiters.end();
      for(list<struct waiter *>::iterator wi=Waiters.begin();
          wi!=wend;){
        struct waiter *w=*wi;

        //DEBUG(LOG(,((void *)w->sens);))
        // found it?
        if(sens==w->sens){

          LOG(ALOG_DEBUG,("waiter: found!"));
            // send data
          if(sens->is_nan)
            snprintf(tmpbuf,MAX_BUF-1,"%ld:g:x\n\n",w->id);
          else
            snprintf(tmpbuf,MAX_BUF-1,"%ld:g:%f\n\n",w->id,sens->value);
          tmpbuf[MAX_BUF-1]='\0';
          Connections[w->conn_n]->sendMessage(tmpbuf);

          // erase waiter and list entry
          {/*log("del w"); */delete w;}
          Waiters.erase(wi);
          break;
        }
        else
          ++wi;
      }

      start=i+1;
      continue;
    }
  }
  mod->sent_ids.clear();
}


void on_module_fail(Ant_agent_module *mod){
  struct waiter *w;

  LOG(ALOG_EXT,("Module %s failed", mod->name));

  // disable module
  mod->active=0;

  // clean up all requests to this module and send answers to servers
  // cycle by waiters
  list<struct waiter *>::iterator wend=Waiters.end();
  for(list<struct waiter *>::iterator wi=Waiters.begin();wi!=wend;){
    w=*wi;

    // found it?
    if(mod==w->sens->module){

      // send answer with bad value
      snprintf(tmpbuf,MAX_BUF-1,"%ld:g:x\n\n",w->id);
      tmpbuf[MAX_BUF-1]='\0';
      Connections[w->conn_n]->sendMessage(tmpbuf);

      // erase waiters entry
      list<struct waiter *>::iterator del=wi++;
      Waiters.erase(del);
    }
    else
      ++wi;
  }
}

void on_server_fail(int i){
  struct waiter *w;

  // clean up waiters
  list<struct waiter *>::iterator wend=Waiters.end();
  for(list<struct waiter *>::iterator wi=Waiters.begin();wi!=wend;){
    w=*wi;

    // found it?
    if(w->conn_n==i){

      LOG(ALOG_EXT,("Drop sensor %s.%s",w->sens->module->name,w->sens->name));
      // erase waiters entry
      list<struct waiter *>::iterator del=wi++;
      Waiters.erase(del);
    }
    else
      ++wi;
  }
}


void daemonize(){
  int pid;

  //if our parent isn't 'init' we do nothing
  if(getppid()!=1) {
    //claim signals ignoring
    signal(SIGTTOU,SIG_IGN);
    signal(SIGTTIN,SIG_IGN);
    signal(SIGTSTP,SIG_IGN);

    //create group and session
    if(pid=fork()){
      //parent

      if(pid<0){
        LOG(ALOG_WARN,("Cannot fork to daemonize"));
      }
      gentle_exit(0);
    }
    //child
    setsid();

    // it is for desctiptors - 0,1,2 - we don't close them
    // but associate with /dev/null in order to prevent
    // situation when the kernel is patched
    int rd=open("/dev/null",O_RDONLY);
    int wr=open("/dev/null",O_WRONLY);
    dup2(rd,0);dup2(wr,1);

    if(log_to_syslog==1){
      dup2(wr,2);
    }
  }
}

#ifdef OLDDD
/*  server socket operations */
/*  this is code from Olivier Aumage, Samuel Thibault */
int build_server_socket(struct addrinfo *ai)
{
    int serv_fd = -1;
    int v = 0;
    int flag = 1;
    char shost[NI_MAXHOST];
    char sport[NI_MAXSERV];
    char host[NI_MAXHOST]="(unknown)";
    char port[NI_MAXSERV]="(null)";

    getnameinfo(ai->ai_addr, ai->ai_addrlen,
                   host, NI_MAXHOST,
                   port, NI_MAXSERV, NI_NUMERICHOST|NI_NUMERICSERV);


    serv_fd = socket(ai->ai_family, ai->ai_socktype, ai->ai_protocol);

    if (serv_fd < 0) {
        LOG(ALOG_WARN | ALOG_NET, ("Cannot create listen socket (%s)",
                                   strerror(errno)));
        return -1;
    }

    /*  this is my addition (Sergey Zhumatiy) */
    setsockopt(serv_fd, SOL_SOCKET, SO_REUSEADDR, &flag, sizeof(int));
    /*  end of addition */

    if (bind(serv_fd, ai->ai_addr, ai->ai_addrlen) < 0) {
        LOG(ALOG_WARN | ALOG_NET, ("Cannot bind to %s:%s (%s)",
            host, port,
            strerror(errno)));
        close(serv_fd);
        return -1;
    }
    if (listen(serv_fd, 1) < 0) {
        LOG(ALOG_WARN | ALOG_NET, ("Cannot listen (%s)", strerror(errno)));
        close(serv_fd);
        return -1;
    }

    /* this is my addition (Sergey Zhumatiy) */
    /* make this socket nonblocked */
    if (fcntl(serv_fd, F_SETFL, O_NONBLOCK |
              fcntl(serv_fd, F_GETFL, O_NONBLOCK)) != 0) {
        LOG(ALOG_WARN | ALOG_NET, ("Cannot make port %s nonblocked (%s)",
                                   sport, strerror(errno)));
        return -1;
    }

    LOG(ALOG_INFO|ALOG_NET, ("Listen on %s:%s",
        host, port));
    return serv_fd;
}

int multi_accept(int *server_s, int nb_server_s,
                 struct sockaddr *addr, socklen_t *len)
{
    int select_server_s = -1;
    int s = -1;
    int i = 0;
    int m = 0;
    int v = 0;


    if (!nb_server_s) {
        LOG(ALOG_WARN | ALOG_NET, ("invalid parameter on accept"));
        return -1;
    }

    fd_set fds;

    FD_ZERO(&fds);

    for (i = 0; i < nb_server_s; i++) {
        if (server_s[i] == -1) {
            nb_server_s--;

            if (i < nb_server_s) {
                server_s[i] = server_s[nb_server_s];
                --i;
                continue;
            }

            break;
        }

        FD_SET(server_s[i], &fds);
        if (m < server_s[i])
            m = server_s[i];
    }

    sleep_time.tv_sec = 0;
    sleep_time.tv_usec = 100;
    v = select(m + 1, &fds, NULL, NULL, &sleep_time);

    if (v < 0) {
        if ((errno != EINTR) && (errno != EALREADY)
            && (errno != EINPROGRESS)) {
            LOG(ALOG_WARN | ALOG_NET, ("Error in select (%s)",
                                       strerror(errno)));
        }
        return -1;
    }

    for (i = 0; i < nb_server_s; i++) {
        if (FD_ISSET(server_s[i], &fds)) {
            select_server_s = server_s[i];
            if ((s = accept(select_server_s, addr, len)) < 0) {
                if ((errno != EINTR) && (errno != EALREADY)
                    && (errno != EINPROGRESS)) {
                    LOG(ALOG_WARN | ALOG_NET, ("Error in accept (%s)",
                                               strerror(errno)));
                    server_s[i] = -1;
                    continue;
                }
            }
            return s;
        }
    }

    return -1;
}

/* end of code from Olivier Aumage, Samuel Thibault */
#endif

void gentle_exit(int code){
    kill(0, SIGQUIT);
    exit(code);
}
