/***************************************************************************
 *
 * This is a part of AntMon project
 *
 *            ant_exe.cpp
 *
 *  Tue Aug 22 14:33:28 2006
 *  Copyright  2006-2007  Sergey Zhumatiy
 *  serg@parallel.ru
 ****************************************************************************/
/*
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include <sys/select.h>
#include <sys/fcntl.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/wait.h>
#include <string.h>
#include <errno.h>
#include <unistd.h>
#include <stdlib.h>

#include <map>
#include <string>
#include "ant_log.hpp"

using namespace std;

int volatile kill_flag=0;

static map < long int, string > exe_hash;
void gentle_exit(int code);

//
//  Execute file with parameters
//  return fd with stdin+stdout associated
//  stderr is redirected specially
//
int ant_exe(const char *exe, char *const *params, int *err_fd)
{
    int id, i;
    int pipe_fd[2];
    int pipe_ed[2];
    //struct timeval tv;

    // create bi-directed pipes
    if (socketpair(AF_UNIX, SOCK_STREAM, PF_UNSPEC, pipe_fd)) {
        LOG(ALOG_WARN, ("Cannot create pipe"));
        return 0;
    }
    if (socketpair(AF_UNIX, SOCK_STREAM, PF_UNSPEC, pipe_ed)) {
        LOG(ALOG_WARN, ("Cannot create pipe"));
        return 0;
    }

    // make it asyncronus
    if (fcntl
        (pipe_fd[0], F_SETFL,
         O_NONBLOCK | fcntl(pipe_fd[0], F_GETFL, O_NONBLOCK)) != 0) {
        LOG(ALOG_WARN,
            ("Cannot make fd %d nonblocked (%s)", pipe_fd[0],
             strerror(errno)));
        return 0;
    }
    if (fcntl
        (pipe_fd[1], F_SETFL,
         O_NONBLOCK | fcntl(pipe_fd[1], F_GETFL, O_NONBLOCK)) != 0) {
        LOG(ALOG_WARN,
            ("Cannot make fd %d nonblocked (%s)", pipe_fd[1],
             strerror(errno)));
        return 0;
    }
    if (fcntl
        (pipe_ed[0], F_SETFL,
         O_NONBLOCK | fcntl(pipe_ed[0], F_GETFL, O_NONBLOCK)) != 0) {
        LOG(ALOG_WARN,
            ("Cannot make fd %d nonblocked (%s)", pipe_ed[0],
             strerror(errno)));
        return 0;
    }
    if (fcntl
        (pipe_ed[1], F_SETFL,
         O_NONBLOCK | fcntl(pipe_ed[1], F_GETFL, O_NONBLOCK)) != 0) {
        LOG(ALOG_WARN,
            ("Cannot make fd %d nonblocked (%s)", pipe_ed[1],
             strerror(errno)));
        return 0;
    }

    id = fork();
    if (id == 0) {              //child - use fd0

        // reset signal handling
        for(i=1;i<255;++i){
            signal(i,SIG_DFL);
        }

        // dup stdin and stdout
        if (dup2(pipe_fd[0], 0) < 0)
            LOG(ALOG_WARN, ("Warning: cannot dup2 0"));
        if (dup2(pipe_fd[0], 1) < 0)
            LOG(ALOG_WARN, ("Warning: cannot dup2 1"));
        if (dup2(pipe_ed[0], 2) < 0)
            LOG(ALOG_WARN, ("Warning: cannot dup2 2"));

/*
        // softly close stderr
//    i=open("/dev/null",O_WRONLY);
        i = open("/tmp/exelog", O_WRONLY | O_APPEND | O_CREAT,
                 S_IRUSR | S_IWUSR);
        if (dup2(i, 2) < 0)
            LOG(ALOG_WARN, ("Warning: cannot dup2 2"));
*/
        // close all other channels
        for (i = 3; i < FD_SETSIZE; ++i) {
            close(i);
        }

        // now there is no reason to write error to log :)
        // make stdin/stdout/stderr asyncronous
        fcntl(0, F_SETFL, fcntl(0, F_GETFL, 0) | O_NONBLOCK);
        fcntl(1, F_SETFL, fcntl(1, F_GETFL, 0) | O_NONBLOCK);
        fcntl(2, F_SETFL, fcntl(2, F_GETFL, 0) | O_NONBLOCK);

        // execute!
        execvp(exe, params);
        gentle_exit(1);
    }

    // error fork?
    if (id < 0) {
        LOG(ALOG_WARN, ("Cannot fork"));
        return 0;
    }

    // parent

    // remember program
    string str = exe;
    for (i = 0; params[i] != NULL; ++i) {
        str += " ";
        str += params[i];
        LOG(ALOG_EXT, ("Executed %ld(%i): %s", (long)id, i, params[i]));
    }
    exe_hash[id] = str;

    // close 'child' part of pipe
    if (close(pipe_fd[0]) < 0)
        LOG(ALOG_WARN, ("Warning: cannot close pipe channel"));

    // make pipes asyncronous
    fcntl(pipe_fd[1], F_SETFL, fcntl(pipe_fd[1], F_GETFL, 0) | O_NONBLOCK);

    if(err_fd!=NULL){
        fcntl(pipe_ed[1], F_SETFL, fcntl(pipe_ed[1], F_GETFL, 0) | O_NONBLOCK);
        *err_fd=pipe_ed[1];
    }
    else{
        i=open("/dev/null", O_WRONLY);
        dup2(pipe_ed[1], i);
    }

    return pipe_fd[1];
}

void exe_sig_chld_handler(int sig)
{
    kill_flag=1;
}

/********************************************
 *
 * kill all zombies...
 *
 * ******************************************/
void take_childs(void){
    int status;
    pid_t pid;
    string exe;

    kill_flag=0;

    for (;;) {
        pid = waitpid(-1, &status, WNOHANG);

        if (pid < 1)
            return;

        exe = exe_hash[pid];
        exe_hash.erase(pid);

        // do this child exist?
        if (exe != "") {
            int sig=0;
            if(WIFSIGNALED(status))
                sig=WTERMSIG(status);
            LOG(ALOG_EXT,
                ("Finished %ld by %d with code %d (%s)", (long int) pid,
                sig, WEXITSTATUS(status), exe.c_str()));
        }
        else{
            LOG(ALOG_EXT,
                ("Finished %ld. Unknown for me. Ignore.", (long int) pid));
        }
    }
}
