/***************************************************************************
 *
 * This is a part of AntMon project
 *
 *            ant_mod.cpp
 *
 *  Tue Aug 22 14:33:28 2006
 *  Copyright  2006-2007  Sergey Zhumatiy
 *  serg@parallel.ru
 ****************************************************************************/
/*
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include <errno.h>
#include <unistd.h>
#include <fcntl.h>
#include <string.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/socket.h>
#include "ant_mod.hpp"
#include "ant_sens.hpp"
#include "ant_log.hpp"
#include "ant_exe.hpp"

using namespace std;

#define DEBUG(x)

#define READ_TIMES_TO_TMOUT 10

struct timeval Ant_agent_module::sleep_time={0,500000};
int Ant_agent_module::send_retr=3;
map<string, Ant_int *> Ant_agent_module::global_ids;
map<string, int *>     Ant_agent_module::global_exe;
map<string, int *>     Ant_agent_module::global_err;
long int               Ant_agent_module::_id=1;
char                   Ant_agent_module::error_buf[ANT_MOD_BUF_CHUNK];

map<long int, void *> Ant_agent_sensor::all;

string Ant_agent_module::modules_path=DEF_EXE_PATH;

//
//  Send request for id
//
int Ant_agent_module::request(Ant_agent_sensor *sens){

  // is it valid?
  if(sens->id==0)
    return 1;

  LOG(ALOG_DEBUG,("mod_request (%s)",sens->name));

  // is it sent already?
  list<Ant_agent_sensor *>::iterator end=sent_ids.end();
  for(list<Ant_agent_sensor *>::iterator li=sent_ids.begin();li!=end;++li){

    // is sent?
    if(*li==sens){
      LOG(ALOG_DEBUG,("mod_request: already sent"));

      return 1;
    }
  }

  // now add request
  sendMessage(sens->name);
  sendMessage("\n");
  sent_ids.push_back(sens);
  return 0;
}

//
//  Write message to buffer
//
void Ant_agent_module::sendMessage(const char *m){
  int len;

  if(!active || !fd)
    return;

  len=strlen(m);
  while(len+out_pos>=out_size){
    // resize buffer!

    out_size+=ANT_MOD_BUF_CHUNK;
    char *buf2=new char[out_size];
    if(buf2==NULL){
      LOG(ALOG_WARN,("Cannot allocate %d bytes of memory.",out_size));
      gentle_exit(0);
    }

    // copy old buffer to new
    memcpy(buf2,out_buffer,out_size-ANT_MOD_BUF_CHUNK);
    {/*log("del m1"); */delete out_buffer;}
    out_buffer=buf2;
  }// resize

  // copy message
  memcpy(out_buffer+out_pos,m,len+1);
  out_pos+=len;
}

//
// Reads modules stderr
//
// Replaces all \n,\r,\0 by '|'
// Always returns pointer to static buffer or NULL
//
const char *Ant_agent_module::readErrors(){
    int pos=0;
    int status;

    if(!ed)
        return NULL;

    // read next char
    while((status=read(ed,error_buf+pos,1))>0){
        if(error_buf[pos]=='\n' ||
           error_buf[pos]=='\r' ||
           error_buf[pos]=='\0')
        {
            error_buf[pos]='|';
        }
        if(++pos>=ANT_MOD_BUF_CHUNK-1){
            error_buf[ANT_MOD_BUF_CHUNK-1]='\0';
            return error_buf;
        }
    }
    if(pos>0)
        return error_buf;

    return NULL;
}

//
// returns 1 if full message is gotten
//
int  Ant_agent_module::readMessage(){
  int status;

  if(!fd || error)
    return 0;

  // read next char
  while((status=read(fd,buffer+pos,1))>0){
    if(++pos>=size){
      // increase buffer
      size+=ANT_MOD_BUF_CHUNK;
      char *buf2=new char[size];
      if(buf2==NULL){
        LOG(ALOG_WARN,("Cannot allocate %d bytes of memory.",size));
        gentle_exit(0);
      }
      memcpy(buf2,buffer,size-ANT_MOD_BUF_CHUNK);
      {/*log("del m2"); */delete buffer;}
      buffer=buf2;
    }

    if(buffer[pos-1]=='\n' && pos>1){
      if(buffer[pos-2]=='\n'){
        buffer[pos-1]='\0';
        //DEBUG(LOG(,(buffer);))
        return 1;
      }
    }
  }

  // error?
  if(status<=0){
    if((errno!=EAGAIN) && (errno!=EINTR)){
      LOG(ALOG_WARN,("Read error (%s)",strerror(errno)));
      error=errno;
      if(pos==0)
        return 2;
    }
  }

  //  DEBUG(LOG(,(buffer);))
  buffer[pos]='\0';
  return 0;
}

void Ant_agent_module::clean(){ // waste current message
  pos=0;
  buffer[0]='\0';
}

int Ant_agent_module::flush(){
  int rest=out_pos;
  int sent,retries;
  struct timeval sl;

  if(!active || !fd)
    return 0;

  if(rest==0)
   return 0;

  retries=send_retr+1;

  if(rest){LOG(ALOG_DEBUG,("writting to %s '%s'...",name, out_buffer+out_pos-rest));}
  while((rest>0) && --retries){
    sent=write(fd,out_buffer+out_pos-rest,rest);

    // error?
    if(sent<=0){
      sent=0;
      if(errno!=EAGAIN && errno!=EINTR){
        LOG(ALOG_EXT,("Write error (%s)",strerror(errno)));
        error=errno;
        break;
      }

      // sleep and try again
      sl=sleep_time;
      select(0,NULL,NULL,NULL,&sl);
    }
    rest-=sent;
  }
  out_pos=0;

  // send final \n
  // !!!!!!!!!!!!!!!!!!!!!!!! NO CHECK !!!!!!!!!!!!!!!!!!!!!
  if(rest==0){
    write(fd,"\n",1);
  }
  // not timed out?
  if(retries)
    return 0;

  // timed out
  LOG(ALOG_EXT,("Cannot send to %s %d times. Disconnect.",name,send_retr));
  finalize();
  return 1;
}

//
//  Argument: executable name with parameters via space
//
void Ant_agent_module::init(const char *c_line){
  // do NOT CARE ABOUT ALREADY INITIALIZED!!!!
  char *pnt[256],*c,*last_zero, prev;
  char com_line[256];
  char reg_buffer[1024];

  int len,i,start,flag, *runned_fd, *runned_ed;
  //int pipe_fd[2];
  struct timeval tv;
  int    read_times=READ_TIMES_TO_TMOUT;

  // is path correct?
  if(c_line[0]=='\0'){
    LOG(ALOG_WARN,("Error! Empty module name. Ignore."));
    fd=0;
    active=0;
    error=0;
    pos=0;
    return;
  }

  strncpy(com_line,c_line,255);
  com_line[255]='\0';
  buffer=new char[ANT_MOD_BUF_CHUNK];
  out_buffer=new char[ANT_MOD_BUF_CHUNK];
  size=out_size=ANT_MOD_BUF_CHUNK;
  pos=out_pos=0;

  // remeber command line
  exe=new char[strlen(com_line)+1];
  strcpy(exe,com_line);

  // check if this exe is runned already
  runned_fd=Ant_agent_module::global_exe[com_line];

  // runned?
  if(runned_fd!=NULL){
    fd=*runned_fd;
    ed=*Ant_agent_module::global_err[com_line];
    active=1;
    error=0;
    pos=0;
    return;
  }


  // find finalizing '\0'
  for(c=com_line;*c;)
    ++c;
  last_zero=c;

  c=com_line;

  // create parameter list
  for(i=0;i<256;){

    //check for eol
    if(*c=='\0')
      break;

    // copy parameter
    pnt[i++]=c;
    while(*c && !isspace(*c))
      ++c;

    //check for eol
    if(*c=='\0')
      break;

    // mark the end
    *c='\0';
    ++c;
  }

  if(i==256){
    LOG(ALOG_WARN,("Warning! Parameter list for module is longer than 256 elements. Shorten it."));
    {/*log("del 002"); */delete pnt[255];}
    i=255;
  }

  pnt[i]=NULL;

  string path=modules_path+":/bin:/usr/bin";
  setenv("PATH",path.c_str(),1);

  // prepare place for stderr
  runned_ed=new int(0);

  // execute action module
  fd=ant_exe(pnt[0], pnt, runned_ed);

  // restore cmd line
  for(c=com_line;c!=last_zero;++c)
    if(*c=='\0')
      *c=' ';

  // error?
  if(fd==0){
    LOG(ALOG_EXT,("Cannot execute module %s (%s)",name,strerror(errno)));
    error=errno;
    active=0;
    delete runned_ed;
    return;
  }

  ed=*runned_ed;
  runned_fd=new int(fd);
  Ant_agent_module::global_exe[com_line]=runned_fd;
  Ant_agent_module::global_err[com_line]=runned_ed;

  active=1;
  error=0;

  // send init string
  write(fd,"list\n",5);

  // read answer
  for(;;){
    prev='\0';
    while(read(fd,buffer+pos,1)>0){
      if(pos>=size-1){
        // increase buffer
        size+=ANT_MOD_BUF_CHUNK;
        char *buf2=new char[size];
        if(buf2==NULL){
          LOG(ALOG_WARN,("Cannot allocate %d bytes of memory.",size));
          gentle_exit(0);
        }
        memcpy(buf2,buffer,size-ANT_MOD_BUF_CHUNK);
        {/*log("del m3"); */delete buffer;}
        buffer=buf2;
      }

      if((buffer[pos]=='\n') && (prev=='\n'))
        break;

      prev=buffer[pos];
      ++pos;
    }

    // successfull read
    if(buffer[pos]=='\n' && (prev=='\n'))
      break;

    if((errno!=EAGAIN) && (errno!=EINTR)){
      LOG(ALOG_EXT,("Read error (%s)",strerror(errno)));
      error=errno;
      active=0;
      return;
    }

    // sleep and check timeout
    if(--read_times==0){
      LOG(ALOG_EXT,("Read timed out"));
      finalize();
      return;
    }

    tv=sleep_time;
    select(0,NULL,NULL,NULL,&tv);
  }

  buffer[pos+1]='\0';
  // List readed. Fill the global_ids!
  LOG(ALOG_DEBUG,("Module capabilities:\n%s",buffer));

  // remember sensor name start.
  start=0;

  // remember module name length
  len=strlen(name);

  if(len>1020){
    LOG(ALOG_WARN,("Warning! module name too long"));
  }
  // read all sensors
  for(i=0;i<=pos;++i){
    if(buffer[i]==';' || buffer[i]=='\n'){
      buffer[i]='\0';

      // generate new id
      _new_id();
      // check if it is already used
      map<string,Ant_int*>::iterator end=global_ids.end();
      flag=1;
      while(flag){
        flag=0;
        for(map<string,Ant_int *>::iterator it=global_ids.begin();
            it!=end;++it){
          pair<string,Ant_int *> pp=*it;

          if(pp.second == NULL){
            LOG(ALOG_WARN,("Strange!!!(%s)",pp.first.c_str()));
            continue;
          }
          // already used
          if(*pp.second == _id){
            flag=1;

            // next check
            _new_id();
            break;
          }
        }
      }
      // now _id is unique

      // add sensor name to global table as 'mod_name.sensor_name'
      strncpy(reg_buffer,name,1023);
      reg_buffer[1023]='\0';
      if(len<1023)
        reg_buffer[len]=' ';
      else{
        len=1023;
        LOG(ALOG_WARN,("Warning! Too long sensor name: %s",reg_buffer));
      }

      // reg_buffer="module_name sensor_name"
      strncpy(reg_buffer+len+1,buffer+start,1022-len);
      Ant_agent_module::global_ids[reg_buffer]=new Ant_int(_id);

      LOG(ALOG_INFO,("Registered sensor '%s' as %ld ...", reg_buffer, _id));

      // Create associated sensor
      Ant_agent_sensor *sens=new Ant_agent_sensor;

      // initialize it
      sens->name=new char[strlen(buffer+start)+1];
      strcpy(sens->name,buffer+start);
      sens->mod_name=new char[strlen(name)+1];
      strcpy(sens->mod_name,name);
      sens->module=this;
      sens->id=_id;

      sens->active=1;
      sens->last_time=0;

      // register it
      Ant_agent_sensor::all[_id]=sens;

      // mark next sensor name start
      start=i+1;
    }
  }// all sensors readed

  pos=0;
}

long int Ant_agent_module::_new_id(){
  if(++_id<1){
    _id=1;
  }
  return _id;
}

void Ant_agent_module::register_sensor(char *sens_name){
  int flag;

  // generate new id
  Ant_agent_module::_new_id();

  // check if it is already used
  map<string,Ant_int*>::iterator end=Ant_agent_module::global_ids.end();
  flag=1;
  while(flag){
    flag=0;
    for(map<string,Ant_int *>::iterator it=Ant_agent_module::global_ids.begin();
        it!=end;++it){

      pair<string,Ant_int *> pp=*it;

      if(pp.second == NULL){
        LOG(ALOG_WARN,("Strange...(%s) [reg_sensor]",pp.first.c_str()));
        continue;
      }
      // already used
      if(*pp.second == _id){
        flag=1;

        // next check
        _new_id();
        break;
      }
    }
  }
  // now _id is unique

  string str=name;
  str+=" ";
  str+=sens_name;
  Ant_agent_module::global_ids[str]=new Ant_int(_id);

  LOG(ALOG_INFO,("Registered sensor '%s' as %ld [reg_sensor]", str.c_str(), _id));

  // Create associated sensor
  Ant_agent_sensor *sens=new Ant_agent_sensor;

  // initialize it
  sens->name=new char[strlen(sens_name)+1];
  strcpy(sens->name,sens_name);
  sens->mod_name=new char[strlen(name)+1];
  strcpy(sens->mod_name,name);
  sens->module=this;
  sens->id=_id;

  sens->active=1;
  sens->last_time=0;

  // register it
  Ant_agent_sensor::all[_id]=sens;
}


void Ant_agent_module::finalize(){
  if(buffer)
    {/*log("del m4"); */delete buffer;}
  if(out_buffer)
    {/*log("del m5"); */delete out_buffer;}
  //  if(name)
  //    delete name;
//  if(exe)
//    {/*log("del m6"); */delete exe;}

  if(fd)
    close(fd);
  if(ed)
    close(ed);

  error=0;
  active=0;
  fd=0;
  size=out_size=pos=out_pos=0;

  buffer=out_buffer=NULL;

  sent_ids.clear();
  Ant_agent_module::global_exe.erase(exe);
  Ant_agent_module::global_err.erase(exe);
}

Ant_agent_module::~Ant_agent_module(){
  finalize();
  if(exe)
    {/*log("del m6"); */delete exe;}
  if(name)
    {/*log("del m9"); */delete name;}
}

Ant_agent_module::Ant_agent_module(const char *n){
  buffer=out_buffer=exe=NULL;
  error=0;
  active=0;
  fd=0;
  size=out_size=pos=out_pos=0;
  next_try=0;

  name=new char [strlen(n)+1];
  strcpy(name,n);
}
