/***************************************************************************
 *
 * This is a part of AntMon project
 *
 *            ant_net.cpp
 *
 *  Tue Jul 31 2007
 *  Copyright  2007  Sergey Zhumatiy
 *  serg@parallel.ru
 ****************************************************************************/
/*
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include <errno.h>
#include <string.h>
#include <fcntl.h>
#include <unistd.h>
#include <stdio.h>
#include "ant_net.hpp"
#include "ant_log.hpp"

static struct timeval sleep_time;

int mk_server_socket(struct addrinfo *ai);

/*  server socket operations */
/*  this is code from Olivier Aumage, Samuel Thibault */
int build_server_socket(unsigned listen_port, int *server_s, int max_serv)
{
    //int serv_fd = -1;
    //int v = 0;
    //int flag = 1;
    //char shost[NI_MAXHOST];
    //char sport[NI_MAXSERV];
    //char host[NI_MAXHOST]="(unknown)";
    char port[NI_MAXSERV];
    struct addrinfo hints;
    struct addrinfo *res = NULL;
    struct addrinfo *ai = NULL;
    int code, nb_server_s;

    nb_server_s=0;
    /* this is code from Olivier Aumage, Samuel Thibault */
    snprintf(port, NI_MAXSERV, "%u", listen_port);

    hints.ai_flags = AI_PASSIVE;
    hints.ai_family = AF_INET; /* PF_UNSPEC; */
    hints.ai_socktype = SOCK_STREAM;
    hints.ai_protocol = IPPROTO_TCP;
    hints.ai_addrlen = 0;
    hints.ai_addr = NULL;
    hints.ai_canonname = NULL;
    hints.ai_next = NULL;

    if ((code = getaddrinfo(NULL, port, &hints, &res))) {
        LOG(ALOG_WARN | ALOG_NET, ("Error in getaddrinfo: %s\n",
                                   gai_strerror(code)));
        if (code == EAI_SYSTEM) {
            LOG(ALOG_WARN | ALOG_NET, ("Details: %s", strerror(errno)));
        }
        return -1;
    }

    for (ai = res; ai && nb_server_s < max_serv; ai = ai->ai_next) {
        int _server_s = -1;

        if (ai->ai_socktype != hints.ai_socktype
            || ai->ai_protocol != hints.ai_protocol)
            continue;

        if ((_server_s = mk_server_socket(ai)) < 0) {
            continue;
        }

        server_s[nb_server_s++] = _server_s;
    }

    freeaddrinfo(res);
    return nb_server_s;
}

int mk_server_socket(struct addrinfo *ai)
{
    int serv_fd = -1;
    //int v = 0;
    int flag = 1;
    //char shost[NI_MAXHOST];
    char sport[NI_MAXSERV];
    char host[NI_MAXHOST]="(unknown)";
    char port[NI_MAXSERV]="(none)";

    getnameinfo(ai->ai_addr, ai->ai_addrlen,
                   host, NI_MAXHOST,
                   port, NI_MAXSERV, NI_NUMERICHOST|NI_NUMERICSERV);


    serv_fd = socket(ai->ai_family, ai->ai_socktype, ai->ai_protocol);

    if (serv_fd < 0) {
        LOG(ALOG_WARN | ALOG_NET, ("Cannot create listen socket (%s)",
                                   strerror(errno)));
        return -1;
    }

    /*  this is my addition (Sergey Zhumatiy) */
    setsockopt(serv_fd, SOL_SOCKET, SO_REUSEADDR, &flag, sizeof(int));
    /*  end of addition */

    if (bind(serv_fd, ai->ai_addr, ai->ai_addrlen) < 0) {
        LOG(ALOG_WARN | ALOG_NET, ("Cannot bind to %s:%s (%s)",
            host, port,
            strerror(errno)));
        close(serv_fd);
        return -1;
    }
    if (listen(serv_fd, 1) < 0) {
        LOG(ALOG_WARN | ALOG_NET, ("Cannot listen (%s)", strerror(errno)));
        close(serv_fd);
        return -1;
    }

    /* this is my addition (Sergey Zhumatiy) */
    /* make this socket nonblocked */
    if (fcntl(serv_fd, F_SETFL, O_NONBLOCK |
              fcntl(serv_fd, F_GETFL, O_NONBLOCK)) != 0) {
        LOG(ALOG_WARN | ALOG_NET, ("Cannot make port %s nonblocked (%s)",
                                   sport, strerror(errno)));
        return -1;
    }

    LOG(ALOG_INFO|ALOG_NET, ("Listen on %s:%s",
        host, port));
    return serv_fd;
}


int get_names_and_port(int serv_fd, char *shost,
                        int numhosts, char *sport)
{
    int code;

    struct sockaddr_storage sas;
    socklen_t sas_len = sizeof(sas);

    if (getsockname(serv_fd, (struct sockaddr *) &sas, &sas_len) < 0) {
        LOG(ALOG_WARN | ALOG_NET, ("Error on getsockname (%s)",
                                   strerror(errno)));
        close(serv_fd);
        return 0;
    }

    if ((code = getnameinfo((struct sockaddr *) &sas, sas_len,
                            shost, NI_MAXHOST, sport, NI_MAXSERV,
                            /*NI_NUMERICSERV |*/NI_NUMERICHOST))) {
        LOG(ALOG_WARN | ALOG_NET, ("Error on getaddrinfo: %s",
                                   gai_strerror(code)));
        if (code == EAI_SYSTEM) {
            LOG(ALOG_WARN | ALOG_NET, ("Details: %s", strerror(errno)));
        }
        close(serv_fd);
        return 0;
    }
    return 0;
}

int multi_accept(int *server_s, int nb_server_s,
                 struct sockaddr *addr, socklen_t *len)
{
    int select_server_s = -1;
    int s = -1;
    int i = 0;
    int m = 0;
    int v = 0;

    *len=sizeof(struct sockaddr);

    if (nb_server_s<1) {
        LOG(ALOG_WARN | ALOG_NET, ("invalid parameter on accept"));
        return -1;
    }

    fd_set fds;

    FD_ZERO(&fds);

    for (i = 0; i < nb_server_s; i++) {
        if (server_s[i] == -1) {
            nb_server_s--;

            if (i < nb_server_s) {
                server_s[i] = server_s[nb_server_s];
                --i;
                continue;
            }

            break;
        }

        FD_SET(server_s[i], &fds);
        if (m < server_s[i])
            m = server_s[i];
    }

    sleep_time.tv_sec = 0;
    sleep_time.tv_usec = 100;
    v = select(m + 1, &fds, NULL, NULL, &sleep_time);

    if (v < 0) {
        if ((errno != EINTR) && (errno != EALREADY)
            && (errno != EINPROGRESS)) {
            LOG(ALOG_WARN | ALOG_NET, ("Error in select (%s)",
                                       strerror(errno)));
        }
        return -1;
    }

    for (i = 0; i < nb_server_s; i++) {
        if (FD_ISSET(server_s[i], &fds)) {
            select_server_s = server_s[i];
            if ((s = accept(select_server_s, addr, len)) < 0) {
                if ((errno != EINTR) && (errno != EALREADY)
                    && (errno != EINPROGRESS)) {
                    LOG(ALOG_WARN | ALOG_NET, ("Error in accept (%s)",
                                               strerror(errno)));
                    server_s[i] = -1;
                    continue;
                }
            }
            return s;
        }
    }

    return -1;
}

/* end of code from Olivier Aumage, Samuel Thibault */
