#ifndef ANT_CONN_HPP
#define ANT_CONN_HPP

#include <string>

#include <netinet/in.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netdb.h>

#define ANT_CONN_BUF_CHUNK 1024
#define ANT_CONN_RETRY_TIME 66

class Ant_connection {
  private:
    struct addrinfo *addr_ptr;  // temporary addr pointer
                                // used for reconnection

  public:
    static struct timeval sleep_time;

    static int send_retr;       // max attempts to send

    struct addrinfo *addr;      // address info
    std::string     name;       // host name
    int port;                   // host port
    int fd;                     // file descriptor

    char connected;             // is connected?
    char enabled;               // is enabled?
    int  error;                 // error code
    char timed_out;             // flag if all params are ALREADY
                                // marked as timed out
    time_t next_connect;        // next reconnecion time
    time_t disconnected_time;   // whan was disconnected

    char *buffer;               // message buffer
    int size;                   // message size
    int pos;                    // current position
    char *out_buffer;           // outgoing message buffer
    int out_size;               // outgoing message size
    int out_pos;                // outgoing message position

    int getMessage(void);       // ret 1 if got full message
    int sendMessage(char *);    // ret 0 always
    int flush(void);            // actually send
    int connect(void);          // do connect
    void disconnect(void);      // abort connection
    void clean(void);           // forget current message

    Ant_connection(void);
    ~Ant_connection(void);
    void finalize(void);        // free memory and disconnect
    void init(const char *);    // initialize all (arg=hostname)
};

#endif
