#include <stdio.h>
#include <fcntl.h>
#include <sys/socket.h>
#include <sys/types.h>
#include <netdb.h>
#include <string.h>
#include <stdlib.h>
#include <errno.h>

/*************************************************

  pop3 check module for Antmon system.

  Returns 0 if service is alive, 1 if not.

  Parameter name is given in form: USER:PASS@HOST
  PASS and USER can be empty strings (':' and '@' cannot me ommitted)

**************************************************/


#define REQ_BUFFER 1023

char req[REQ_BUFFER+1];


#if 0
int GetAddress(struct sockaddr_in *addr,const char *host, short int port){
  struct addrinfo hint,*res;
  struct sockaddr_in *in;
  int code;
  hint.ai_family=PF_INET;
  hint.ai_flags=0;
  hint.ai_socktype=0;
  hint.ai_protocol=0;
  hint.ai_next=NULL;
  hint.ai_addr=NULL;
  hint.ai_addrlen=0;

  bzero(addr, sizeof(*addr));
  addr->sin_family = AF_INET;
  addr->sin_port = htons(port);
  if ( host != 0 ) {
    if(code=getaddrinfo(host, NULL,&hint,&res)){
      /* ("Cannot resolve %s (%s)",host,gai_strerror(code)); */
      return -1;
    }

    in=(struct sockaddr_in *)&(res->ai_addr);
    addr->sin_addr=in->sin_addr;
    /* ("Resolved %s",host); */
    freeaddrinfo(res);
    return 0;
  }
  addr->sin_addr.s_addr = htonl(INADDR_ANY);
  return 0;
}
#else

struct in_addr *atoaddr(const char * address)
{
        struct hostent *host;
        static struct in_addr saddr;

        saddr.s_addr = inet_addr(address);
        if(saddr.s_addr != -1)
                return &saddr;

        host = gethostbyname(address);
        if(host != NULL)
                return (struct in_addr *) *host->h_addr_list;

        return NULL;
}

int GetAddress(struct sockaddr_in *address, const char *host, int port){
	/* First convert service from a string, to a number... */
	struct in_addr *addr;

	addr = atoaddr(host);
	if(addr == NULL){
          /* ("Invalid network adress, can't make connection"); */
          /*fprintf(stderr,"Cannot resolve %s\n",host); /*debug*/
          return 1;
        }

	bzero((char *)address,sizeof(*address));
	address->sin_family = AF_INET;
	address->sin_port = htons(port);
	address->sin_addr.s_addr = addr->s_addr;
        return 0;
}

#endif

int readstring(int fd){
  char firstchar,c;
  read(fd,&firstchar,1);

  c=firstchar;
  while(c!='\n')
    read(fd,&c,1);

  return firstchar;
}

/*
   Real pop3 check.

*/
int pop3_check(const char *user, const char *pass, const char *host){
  struct sockaddr_in addr;
  int fd;
  char buffer[1024];

  if(GetAddress(&addr,host,110)!=0){
    /* error */
    return 1;
  }

  if((fd=socket(PF_INET,SOCK_STREAM,0)) <0 ) {
    /*fprintf(stderr,"Error creating socket: %s",strerror(errno));/*debug*/
    return 1;
  }
  if(connect(fd,(const struct sockaddr *)&addr,sizeof(struct sockaddr_in))<0){
    /*fprintf(stderr,"Cannot connect to %s (%s)",host,strerror(errno));/*debug*/
    return 1;
  }

  if(readstring(fd)=='+'){
    if(strlen(user)>0){
      write(fd,"USER ",5);
      write(fd,user,strlen(user));
      write(fd,"\n",1);
      if(readstring(fd)=='+'){
        write(fd,"PASS ",5);
        write(fd,pass,strlen(pass));
        write(fd,"\n",1);
        if(readstring(fd)=='+'){
          write(fd,"QUIT\n",5);
          readstring(fd);
          close(fd);
          return 0;
          /*fprintf(stderr,"OK! 1\n");/*debug*/
          
        }
      }
    }
    else{ /* do not try to check account */
      write(fd,"QUIT\n",5);
      readstring(fd);
      close(fd);
      /*fprintf(stderr,"OK! 2\n");/*debug*/
      return 0;
    }
  }
  write(fd,"QUIT\n",5);
  readstring(fd);
  close(fd);
  /*fprintf(stderr,"FAIL! 1\n");/*debug*/
  return 1;
}


/*  Does the actual check and writes the result.
    In 'req' all requests are stored.
    Result must be written on EVERY request.
*/
void do_check(){
  int i=0,begin=0;
  int user,pass,host;
  int err;

  /* check every request */
  for(err=0;err==0;){
    while(req[i]!='\0')
      ++i;

    /* empty request - end of buffer */
    if(i==begin)
      break;

    /* req+begin = request string */
    /*                            */
    /* format: user:pass@host     */
    /*                            */
    /* user and pass may be null  */

    user=begin;
    for(pass=begin;req[pass]!=':';++pass)
      if(req[pass]=='\0'){
        /* Error! Bad request! */
        write(1,"0\n",2);
        err=1;
        break;
      }
    if(err)
      break;

    /* pass -> ':' */
    req[pass]='\0';
    ++pass;

    for(host=pass;req[host]!='@';++host)
      if(req[host]=='\0'){
        /* Error! Bad request! */
        write(1,"0\n",2);
        err=1;
        break;
      }
    if(err)
      break;
    
    /* host -> '@' */
    req[host]='\0';
    ++host;

    if(pop3_check(req+user,req+pass,req+host)==0){
      /* success */
      write(1,"1\n",2);
    }
    else{
      /* fail */
      write(1,"0\n",2);
    }
    begin=++i;
  }
  write(1,"\n",1);
  return;
}

/*  main loop  */
int main(){
  char c,prev;
  int flag,i;
  struct timeval tv;


  /* important !!! */
  fcntl(0,F_SETFL,fcntl(0,F_GETFL,0)|O_NONBLOCK);
  fcntl(1,F_SETFL,fcntl(1,F_GETFL,0)|O_NONBLOCK);

  /* read first line and ignore it. */
  c='\0';
  while(c!='\n'){
    if(read(0,&c,1)<1){
      tv.tv_sec=0;
      tv.tv_usec=500000;
      select(0,NULL,NULL,NULL,&tv);
    }
  }

  /* print list of supported sensors via space */

  write(1,"*\n",2);

  /* infinite loop */
  for(;;){

    i=0;
    flag=0;

    /* store here last readed character */
    prev='\0';

    /* read request */
    for(;;){

      if(i>REQ_BUFFER){
        /*fprintf(stderr,"BAD request! Too long!\n");/*debug*/
        exit(1);
      }

      /*one-by-one byte*/
      while(read(0,&c,1)>0){

        /* end of sensor name? */
        if(c=='\n'){

          /* end of whole request */
          if(prev=='\n'){
            flag=1;
            req[i++]='\0';
            break;
          }

          /* mark the end of request string */
          req[i++]='\0';
          prev=c;
          continue;
        }
        prev=c;
        req[i++]=c;
      }
      /* end of whole request? */
      if(flag)
        break;

      /* sleep 0,5 seconds */
      tv.tv_sec=0;
      tv.tv_usec=500000;
      select(0,NULL,NULL,NULL,&tv);
    }

    req[i]='\0';
    do_check();
  }
  exit(0);
}
