/**************************************************

  AntMon module syssstat

  This module gets information about:

      cpu(s)/loadavg
      memory usage
      swap/paging
      inet interfaces
      nfs client/server usage

***************************************************/

/*
    BUGS:
      - if system has NO network interfaces, module will fail
      - no checks for success files open is performed
      - no mallock success checks performed
*/


#define _XOPEN_SOURCE 500
#include <stdio.h>
#include <stdarg.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <sys/time.h>
#include <fcntl.h>
#include <sys/types.h>
#include <unistd.h>
#include <locale.h>

#define BUF_SIZE     1024
#define IF_NAME_LEN  16
#define MAX_REQ_LINE 8192

int debug=0;  /* use debug */


int stat_long=0; /* 8 points in cpu stat string */
int k26=0;       /* kernel version is 2.6.x or later */
int stat26_fd,
  stat_fd,
  loadavg_fd,
  mem_fd,
  nfs_fd,
  nfsd_fd,
  net_fd;

long int diff; /* time differecne in milliseconds */

int num_cpu, num_if;

struct cpu_stat{
  double user;
  double nice;
  double system;      /* aggregated with iowait,irq,etc*/
  double idle;
  double system_real;
  double iowait;
  double irq;
  double softirq;
  double steal;
};

struct main_stat{
  double pgpin;
  double pgpout;
  double swpin;
  double swpout;
  struct cpu_stat all_cpu;
  struct cpu_stat *cpu;

  double context;
  double forks;
  double mem_total;
  double mem_free;
  double mem_shared;
  double mem_buff;
  double mem_cached;
  double mem_swp;
  double mem_swp_free;
};

struct netstat{
  char if_name[IF_NAME_LEN];
  double rcv_bytes;
  double rcv_pckts;
  double rcv_errs;
  double rcv_drop;
  double rcv_fifo;
  double rcv_frame;
  double rcv_compr;
  double rcv_multi;

  double sent_bytes;
  double sent_pckts;
  double sent_errs;
  double sent_drop;
  double sent_fifo;
  double sent_colls;
  double sent_carr;
  double sent_compr;
};

struct nfs_stat{
  double nfs_count;
  double nfs_retrans;
  double nfs_auth;
  double nfs2_null;
  double nfs2_getattr;
  double nfs2_setattr;
  double nfs2_root;
  double nfs2_lookup;
  double nfs2_readlink;
  double nfs2_read;
  double nfs2_writechache;
  double nfs2_write;
  double nfs2_create;
  double nfs2_remove;
  double nfs2_rename;
  double nfs2_link;
  double nfs2_symlink;
  double nfs2_mkdir;
  double nfs2_rmdir;
  double nfs2_readdir;
  double nfs2_statfs;
  double nfs3_null;
  double nfs3_getattr;
  double nfs3_setattr;
  double nfs3_lookup;
  double nfs3_access;
  double nfs3_readlink;
  double nfs3_read;
  double nfs3_write;
  double nfs3_create;
  double nfs3_mkdir;
  double nfs3_symlink;
  double nfs3_mknod;
  double nfs3_remove;
  double nfs3_rmdir;
  double nfs3_rename;
  double nfs3_link;
  double nfs3_readdir;
  double nfs3_readdirplus;
  double nfs3_fsstat;
  double nfs3_fsinfo;
  double nfs3_pathconf;
  double nfs3_commit;
};

struct nfsd_stat{
  double nfsd_calls;
  double nfsd_badcalls;
  double nfsd_badauth;
  double nfsd_badclnt;
  double nfsd_xdrcall;
  double nfsd2_null;
  double nfsd2_getattr;
  double nfsd2_setattr;
  double nfsd2_root;
  double nfsd2_lookup;
  double nfsd2_readlink;
  double nfsd2_read;
  double nfsd2_writechache;
  double nfsd2_write;
  double nfsd2_create;
  double nfsd2_remove;
  double nfsd2_rename;
  double nfsd2_link;
  double nfsd2_symlink;
  double nfsd2_mkdir;
  double nfsd2_rmdir;
  double nfsd2_readdir;
  double nfsd2_statfs;
  double nfsd3_null;
  double nfsd3_getattr;
  double nfsd3_setattr;
  double nfsd3_lookup;
  double nfsd3_access;
  double nfsd3_readlink;
  double nfsd3_read;
  double nfsd3_write;
  double nfsd3_create;
  double nfsd3_mkdir;
  double nfsd3_symlink;
  double nfsd3_mknod;
  double nfsd3_remove;
  double nfsd3_rmdir;
  double nfsd3_rename;
  double nfsd3_link;
  double nfsd3_readdir;
  double nfsd3_readdirplus;
  double nfsd3_fsstat;
  double nfsd3_fsinfo;
  double nfsd3_pathconf;
  double nfsd3_commit;

  double nfsd3_cache;
  double nfsd3_miss;
  double nfsd3_nocache;
  double nfsd3_io_read;
  double nfsd3_io_write;
  double nfsd3_netcount;
  double nfsd3_thread_sec;
  double nfsd3_thread_10;
  double nfsd3_thread_20;
  double nfsd3_thread_30;
  double nfsd3_thread_40;
  double nfsd3_thread_50;
  double nfsd3_thread_60;
  double nfsd3_thread_70;
  double nfsd3_thread_80;
  double nfsd3_thread_90;
  double nfsd3_thread_100;
};
struct main_stat  stat_new, stat_last, stat_cur;
double            loadavg1_cur, loadavg5_cur, loadavg15_cur;
struct netstat    *netstat_cur, *netstat_new, *netstat_last;
struct nfs_stat   nfsstat_last,nfsstat_cur,nfsstat_new;
struct nfsd_stat  nfsdstat_last,nfsdstat_cur,nfsdstat_new;



struct str_liptr{
  char *str;
  double *ptr;
};
/*    table for values search */
struct str_liptr baselist[]={
  {"cpu_user",&stat_cur.all_cpu.user},
  {"cpu_nice",&stat_cur.all_cpu.nice},
  {"cpu_system",&stat_cur.all_cpu.system},
  {"cpu_idle",&stat_cur.all_cpu.idle},
  {"cpu_sys_real",&stat_cur.all_cpu.system_real},
  {"cpu_iowait",&stat_cur.all_cpu.iowait},
  {"cpu_irq",&stat_cur.all_cpu.irq},
  {"cpu_softirq",&stat_cur.all_cpu.softirq},
  {"cpu_steal",&stat_cur.all_cpu.steal},
  {"pgpin",&stat_cur.pgpin},
  {"pgpout",&stat_cur.pgpout},
  {"swpin",&stat_cur.swpin},
  {"swpout",&stat_cur.swpout},
  {"context",&stat_cur.context},
  {"forks",&stat_cur.forks},
  {"mem_total",&stat_cur.mem_total},
  {"mem_free",&stat_cur.mem_free},
  {"mem_shared",&stat_cur.mem_shared},
  {"mem_buff",&stat_cur.mem_buff},
  {"mem_cached",&stat_cur.mem_cached},
  {"mem_swp",&stat_cur.mem_swp},
  {"mem_swp_free",&stat_cur.mem_swp_free},
  {"nfs_count",&nfsstat_cur.nfs_count},
  {"nfs_retrans",&nfsstat_cur.nfs_retrans},
  {"nfs_auth",&nfsstat_cur.nfs_auth},
  {"nfs2_null",&nfsstat_cur.nfs2_null},
  {"nfs2_getattr",&nfsstat_cur.nfs2_getattr},
  {"nfs2_setattr",&nfsstat_cur.nfs2_setattr},
  {"nfs2_root",&nfsstat_cur.nfs2_root},
  {"nfs2_lookup",&nfsstat_cur.nfs2_lookup},
  {"nfs2_readlink",&nfsstat_cur.nfs2_readlink},
  {"nfs2_read",&nfsstat_cur.nfs2_read},
  {"nfs2_writechache",&nfsstat_cur.nfs2_writechache},
  {"nfs2_write",&nfsstat_cur.nfs2_write},
  {"nfs2_create",&nfsstat_cur.nfs2_create},
  {"nfs2_remove",&nfsstat_cur.nfs2_remove},
  {"nfs2_rename",&nfsstat_cur.nfs2_rename},
  {"nfs2_link",&nfsstat_cur.nfs2_link},
  {"nfs2_symlink",&nfsstat_cur.nfs2_symlink},
  {"nfs2_mkdir",&nfsstat_cur.nfs2_mkdir},
  {"nfs2_rmdir",&nfsstat_cur.nfs2_rmdir},
  {"nfs2_readdir",&nfsstat_cur.nfs2_readdir},
  {"nfs2_statfs",&nfsstat_cur.nfs2_statfs},
  {"nfs3_null",&nfsstat_cur.nfs3_null},
  {"nfs3_getattr",&nfsstat_cur.nfs3_getattr},
  {"nfs3_setattr",&nfsstat_cur.nfs3_setattr},
  {"nfs3_lookup",&nfsstat_cur.nfs3_lookup},
  {"nfs3_access",&nfsstat_cur.nfs3_access},
  {"nfs3_readlink",&nfsstat_cur.nfs3_readlink},
  {"nfs3_read",&nfsstat_cur.nfs3_read},
  {"nfs3_write",&nfsstat_cur.nfs3_write},
  {"nfs3_create",&nfsstat_cur.nfs3_create},
  {"nfs3_mkdir",&nfsstat_cur.nfs3_mkdir},
  {"nfs3_symlink",&nfsstat_cur.nfs3_symlink},
  {"nfs3_mknod",&nfsstat_cur.nfs3_mknod},
  {"nfs3_remove",&nfsstat_cur.nfs3_remove},
  {"nfs3_rmdir",&nfsstat_cur.nfs3_rmdir},
  {"nfs3_rename",&nfsstat_cur.nfs3_rename},
  {"nfs3_link",&nfsstat_cur.nfs3_link},
  {"nfs3_readdir",&nfsstat_cur.nfs3_readdir},
  {"nfs3_readdirplus",&nfsstat_cur.nfs3_readdirplus},
  {"nfs3_fsstat",&nfsstat_cur.nfs3_fsstat},
  {"nfs3_fsinfo",&nfsstat_cur.nfs3_fsinfo},
  {"nfs3_pathconf",&nfsstat_cur.nfs3_pathconf},
  {"nfs3_commit",&nfsstat_cur.nfs3_commit},
  {"nfsd_calls",&nfsdstat_cur.nfsd_calls},
  {"nfsd_badcalls",&nfsdstat_cur.nfsd_badcalls},
  {"nfsd_badauth",&nfsdstat_cur.nfsd_badauth},
  {"nfsd_badclnt",&nfsdstat_cur.nfsd_badclnt},
  {"nfsd_xdrcall",&nfsdstat_cur.nfsd_xdrcall},
  {"nfsd2_null",&nfsdstat_cur.nfsd2_null},
  {"nfsd2_getattr",&nfsdstat_cur.nfsd2_getattr},
  {"nfsd2_setattr",&nfsdstat_cur.nfsd2_setattr},
  {"nfsd2_root",&nfsdstat_cur.nfsd2_root},
  {"nfsd2_lookup",&nfsdstat_cur.nfsd2_lookup},
  {"nfsd2_readlink",&nfsdstat_cur.nfsd2_readlink},
  {"nfsd2_read",&nfsdstat_cur.nfsd2_read},
  {"nfsd2_writechache",&nfsdstat_cur.nfsd2_writechache},
  {"nfsd2_write",&nfsdstat_cur.nfsd2_write},
  {"nfsd2_create",&nfsdstat_cur.nfsd2_create},
  {"nfsd2_remove",&nfsdstat_cur.nfsd2_remove},
  {"nfsd2_rename",&nfsdstat_cur.nfsd2_rename},
  {"nfsd2_link",&nfsdstat_cur.nfsd2_link},
  {"nfsd2_symlink",&nfsdstat_cur.nfsd2_symlink},
  {"nfsd2_mkdir",&nfsdstat_cur.nfsd2_mkdir},
  {"nfsd2_rmdir",&nfsdstat_cur.nfsd2_rmdir},
  {"nfsd2_readdir",&nfsdstat_cur.nfsd2_readdir},
  {"nfsd2_statfs",&nfsdstat_cur.nfsd2_statfs},
  {"nfsd3_null",&nfsdstat_cur.nfsd3_null},
  {"nfsd3_getattr",&nfsdstat_cur.nfsd3_getattr},
  {"nfsd3_setattr",&nfsdstat_cur.nfsd3_setattr},
  {"nfsd3_lookup",&nfsdstat_cur.nfsd3_lookup},
  {"nfsd3_access",&nfsdstat_cur.nfsd3_access},
  {"nfsd3_readlink",&nfsdstat_cur.nfsd3_readlink},
  {"nfsd3_read",&nfsdstat_cur.nfsd3_read},
  {"nfsd3_write",&nfsdstat_cur.nfsd3_write},
  {"nfsd3_create",&nfsdstat_cur.nfsd3_create},
  {"nfsd3_mkdir",&nfsdstat_cur.nfsd3_mkdir},
  {"nfsd3_symlink",&nfsdstat_cur.nfsd3_symlink},
  {"nfsd3_mknod",&nfsdstat_cur.nfsd3_mknod},
  {"nfsd3_remove",&nfsdstat_cur.nfsd3_remove},
  {"nfsd3_rmdir",&nfsdstat_cur.nfsd3_rmdir},
  {"nfsd3_rename",&nfsdstat_cur.nfsd3_rename},
  {"nfsd3_link",&nfsdstat_cur.nfsd3_link},
  {"nfsd3_readdir",&nfsdstat_cur.nfsd3_readdir},
  {"nfsd3_readdirplus",&nfsdstat_cur.nfsd3_readdirplus},
  {"nfsd3_fsstat",&nfsdstat_cur.nfsd3_fsstat},
  {"nfsd3_fsinfo",&nfsdstat_cur.nfsd3_fsinfo},
  {"nfsd3_pathconf",&nfsdstat_cur.nfsd3_pathconf},
  {"nfsd3_commit",&nfsdstat_cur.nfsd3_commit},
  {"nfsd3_cache",&nfsdstat_cur.nfsd3_cache},
  {"nfsd3_miss",&nfsdstat_cur.nfsd3_miss},
  {"nfsd3_nocache",&nfsdstat_cur.nfsd3_nocache},
  {"nfsd3_io_read",&nfsdstat_cur.nfsd3_io_read},
  {"nfsd3_io_write",&nfsdstat_cur.nfsd3_io_write},
  {"nfsd3_netcount",&nfsdstat_cur.nfsd3_netcount},
  {"nfsd3_thread_sec",&nfsdstat_cur.nfsd3_thread_sec},
  {"nfsd3_thread_10",&nfsdstat_cur.nfsd3_thread_10},
  {"nfsd3_thread_20",&nfsdstat_cur.nfsd3_thread_20},
  {"nfsd3_thread_30",&nfsdstat_cur.nfsd3_thread_30},
  {"nfsd3_thread_40",&nfsdstat_cur.nfsd3_thread_40},
  {"nfsd3_thread_50",&nfsdstat_cur.nfsd3_thread_50},
  {"nfsd3_thread_60",&nfsdstat_cur.nfsd3_thread_60},
  {"nfsd3_thread_70",&nfsdstat_cur.nfsd3_thread_70},
  {"nfsd3_thread_80",&nfsdstat_cur.nfsd3_thread_80},
  {"nfsd3_thread_90",&nfsdstat_cur.nfsd3_thread_90},
  {"nfsd3_thread_100",&nfsdstat_cur.nfsd3_thread_100},
  {"loadavg1", &loadavg1_cur},
  {"loadavg5", &loadavg5_cur},
  {"loadavg15", &loadavg15_cur},
  {NULL,NULL}
};

struct str_liptr *ifaceslist;


/* print string in async mode. */
void print(const char *s,...){
  static char buffer[BUF_SIZE];
  int len,written,shift;
  va_list list;

  if(s==NULL)
    return;

  va_start(list,s);
  vsnprintf(buffer,BUF_SIZE-1,s,list);
  va_end(list);

  buffer[BUF_SIZE-1]='\0';

  if(debug){
    if(strcmp(buffer,"\n")) /* do not log only eol */
      fprintf(stderr,"SS-> '%s'\n",buffer);
  }

  len=strlen(buffer);
  shift=0;
  while(shift<len){
    written=write(0,buffer+shift,len-shift);

    /* was there an error? */
    if(written<0){
      if((errno==EAGAIN)||(errno==EINTR)){

        /* just sleep and try again */
        usleep(100);
      }
      else{
        exit(1);
      }
    }
    else{ /* all ok */
      shift+=written;
    }
  }
}

/* reads a string in asyncronous mode */
char *getstr(int fd){
  static char buffer[BUF_SIZE];

  int shift=0,res;
  char c;
  for(;;){
    res=read(fd,&c,1);
    if(res==0){ /* eof */
      if(shift>0)
        break;
      /* eof at first character - error! */
      return NULL;
    }
    if((res<0) && (errno==EAGAIN)){
      /* just sleep and try again */
      usleep(100);
      continue;
    }

    if(res<0){ /* error or eof */
      if(shift)
        break;
      return NULL;
    }

    /* add to buffer */
    buffer[shift]=c;
    if((c=='\0') || (c=='\n'))
      break;
    if(++shift>BUF_SIZE-2)
      break;
  }
  buffer[shift]='\0';
  return buffer;
}


/* --------------------------------------------
     reads cpu and system stats into
           stat_cur

     returns 1 if error
             0 if ok
   -------------------------------------------- */
int get_info_stat(void){
  char *buf;
  int i,readed,n;
  static struct timeval curtime, lasttime={0,0};


  /* try to get info from 2.6 version kernel */
  if(k26){
    if(lseek(stat26_fd,0,SEEK_SET)!=0){
      /* reopen file */
      close(stat26_fd);
      if((stat26_fd=open("/proc/vmstat",O_RDONLY))<0){
        print("*Cannot open /proc/vmstat.");
        return 1;
      }
    }

    readed=0;
    while((buf=getstr(stat26_fd))!=NULL && (readed<4)){
      if(strncmp("pgpgin ",buf,7)==0){
        for(i=7;buf[i]==' ';)
          ++i;
        sscanf(buf+i,"%lg",&stat_new.pgpin);
        ++readed;
        continue;
      }
      if(strncmp("pgpgout ",buf,8)==0){
        for(i=8;buf[i]==' ';)
          ++i;
        sscanf(buf+i,"%lg",&stat_new.pgpout);
        ++readed;
        continue;
      }
      if(strncmp("pswpin ",buf,7)==0){
        for(i=7;buf[i]==' ';)
          ++i;
        sscanf(buf+i,"%lg",&stat_new.swpin);
        ++readed;
        continue;
      }
      if(strncmp("pswpout ",buf,8)==0){
        for(i=8;buf[i]==' ';)
          ++i;
        sscanf(buf+i,"%lg",&stat_new.swpout);
        ++readed;
        continue;
      }
    }
  }

  /* now get info from 2.4-style kernel */
  if(lseek(stat_fd,0,SEEK_SET)!=0){
    /* reopen file */
    close(stat_fd);
    if((stat_fd=open("/proc/stat",O_RDONLY))<0){
      print("*Cannot open /proc/stat.");
      exit(1);
    }
  }

  /* line-by-line */
  readed=0;
  while((buf=getstr(stat_fd))!=NULL && (readed<5+num_cpu)){

    /* pgpin/pgpout */
    if(strncmp("page ",buf,5)==0){
      for(i=5;buf[i]==' ';)
        ++i;
      sscanf(buf+i,"%lg %lg",&stat_new.pgpin,&stat_new.pgpout);
      ++readed;
      continue;
    }
    /* swpin/swpout */
    if(strncmp("swap ",buf,5)==0){
      for(i=5;buf[i]==' ';)
        ++i;
      sscanf(buf+i,"%lg %lg",&stat_new.swpin,&stat_new.swpout);
      ++readed;
      continue;
    }
    /* cpustat  */
    if(strncmp("cpu ",buf,4)==0){
      for(i=4;buf[i]==' ';)
        ++i;
      if(stat_long){
        sscanf(buf+i,"%lg %lg %lg %lg %lg %lg %lg %lg",
               &stat_new.all_cpu.user,
               &stat_new.all_cpu.nice,
               &stat_new.all_cpu.system_real,
               &stat_new.all_cpu.idle,
               &stat_new.all_cpu.iowait,
               &stat_new.all_cpu.irq,
               &stat_new.all_cpu.softirq,
               &stat_new.all_cpu.steal);
        stat_new.all_cpu.user    /= num_cpu;
        stat_new.all_cpu.nice    /= num_cpu;
        stat_new.all_cpu.system_real /= num_cpu;
        stat_new.all_cpu.idle    /= num_cpu;
        stat_new.all_cpu.iowait  /= num_cpu;
        stat_new.all_cpu.irq     /= num_cpu;
        stat_new.all_cpu.softirq /= num_cpu;
        stat_new.all_cpu.steal   /= num_cpu;
        stat_new.all_cpu.system=
          stat_new.all_cpu.system_real +
          stat_new.all_cpu.iowait +
          stat_new.all_cpu.irq    +
          stat_new.all_cpu.softirq+
          stat_new.all_cpu.steal;
/*
        if(debug){fprint(stderr,"user=%lf nice=%lf ssystem=%lf idle=%lf\nwait=%lf irq=%lf softirq=%lf steal=%lf\n",
                        stat_new.all_cpu.user,
                        stat_new.all_cpu.nice,
                        stat_new.all_cpu.system_real,
                        stat_new.all_cpu.idle,
                        stat_new.all_cpu.iowait,
                        stat_new.all_cpu.irq,
                        stat_new.all_cpu.softirq,
                        stat_new.all_cpu.steal);
        }
*/
      }
      else{
        sscanf(buf+i,"%lg %lg %lg %lg",
               &stat_new.all_cpu.user,
               &stat_new.all_cpu.nice,
               &stat_new.all_cpu.system,
               &stat_new.all_cpu.idle);
        stat_new.all_cpu.user   /= num_cpu;
        stat_new.all_cpu.nice   /= num_cpu;
        stat_new.all_cpu.system /= num_cpu;
        stat_new.all_cpu.idle   /= num_cpu;
        stat_new.all_cpu.system_real=stat_new.all_cpu.system;
        stat_new.all_cpu.irq=
          stat_new.all_cpu.softirq=
          stat_new.all_cpu.iowait=
          stat_new.all_cpu.steal=0;
      }
      ++readed;
      continue;
    }
    /* cpustat-per-cpu  */
    if(strncmp("cpu",buf,3)==0){
      n=0;

      /* get cpu number */
      for(i=3;(buf[i]>='0') && (buf[i]<='9');){
        n=n*10+buf[i]-'0';
        ++i;
      }
      if(n>=num_cpu)
        continue;

      while(buf[i]==' ')
        ++i;
      if(stat_long){
        sscanf(buf+i,"%lg %lg %lg %lg %lg %lg %lg %lg",
               &stat_new.cpu[n].user,
               &stat_new.cpu[n].nice,
               &stat_new.cpu[n].system_real,
               &stat_new.cpu[n].idle,
               &stat_new.cpu[n].iowait,
               &stat_new.cpu[n].irq,
               &stat_new.cpu[n].softirq,
               &stat_new.cpu[n].steal);
        stat_new.cpu[n].system=
          stat_new.cpu[n].system_real +
          stat_new.cpu[n].iowait +
          stat_new.cpu[n].irq    +
          stat_new.cpu[n].softirq+
          stat_new.cpu[n].steal;
      }
      else{
        sscanf(buf+i,"%lg %lg %lg %lg",
               &stat_new.cpu[n].user,
               &stat_new.cpu[n].nice,
               &stat_new.cpu[n].system,
               &stat_new.cpu[n].idle);
        stat_new.cpu[n].system_real=stat_new.cpu[n].system;
        stat_new.cpu[n].irq=
          stat_new.cpu[n].softirq=
          stat_new.cpu[n].iowait=
          stat_new.cpu[n].steal=0;
      }
      ++readed;
      continue;
    }
    /* context switches  */
    if(strncmp("ctxt ",buf,5)==0){
      for(i=5;buf[i]==' ';)
        ++i;
      sscanf(buf+i,"%lg",&stat_new.context);
      ++readed;
      continue;
    }
    /* forks  */
    if(strncmp("processes ",buf,10)==0){
      for(i=10;buf[i]==' ';)
        ++i;
      sscanf(buf+i,"%lg",&stat_new.forks);
      ++readed;
      continue;
    }
/*
    elsif(m{^disk_io:\s}){
      while(m{\((\d+),(\d+)\):\((\d+),(\d+),(\d+),(\d+),(\d+)\)}g){
        $info_stat_new{"disk_$1_$2_r"}=$4;
        $info_stat_new{"disk_$1_$2_w"}=$6;
      }
    }
*/
  } /* while(getstr)*/


  gettimeofday(&curtime,NULL);

  /* count time interval in milliseconds */
  diff=(curtime.tv_sec-lasttime.tv_sec)*1000+(curtime.tv_usec-lasttime.tv_usec)/1000;

  if(diff==0)
    diff=1000;


  /* count increments */
  stat_cur.pgpin  =((stat_new.pgpin  -stat_last.pgpin  )*1000.0)/diff;
  stat_cur.pgpout =((stat_new.pgpout -stat_last.pgpout )*1000.0)/diff;
  stat_cur.swpin  =((stat_new.swpin  -stat_last.swpin  )*1000.0)/diff;
  stat_cur.swpout =((stat_new.swpout -stat_last.swpout )*1000.0)/diff;
  stat_cur.context=((stat_new.context-stat_last.context)*1000.0)/diff;
  stat_cur.forks  =((stat_new.forks  -stat_last.forks  )*1000.0)/diff;
  stat_cur.all_cpu.user  =((stat_new.all_cpu.user  -stat_last.all_cpu.user  )*1000.0)/diff;
  stat_cur.all_cpu.nice  =((stat_new.all_cpu.nice  -stat_last.all_cpu.nice  )*1000.0)/diff;
  stat_cur.all_cpu.system=((stat_new.all_cpu.system-stat_last.all_cpu.system)*1000.0)/diff;
  stat_cur.all_cpu.idle  =((stat_new.all_cpu.idle  -stat_last.all_cpu.idle  )*1000.0)/diff;
  stat_cur.all_cpu.system_real=((stat_new.all_cpu.system_real-stat_last.all_cpu.system_real)*1000.0)/diff;
  stat_cur.all_cpu.irq   =((stat_new.all_cpu.irq   -stat_last.all_cpu.irq   )*1000.0)/diff;
  stat_cur.all_cpu.softirq=((stat_new.all_cpu.softirq-stat_last.all_cpu.softirq)*1000.0)/diff;
  stat_cur.all_cpu.iowait=((stat_new.all_cpu.iowait-stat_last.all_cpu.iowait)*1000.0)/diff;
  stat_cur.all_cpu.steal =((stat_new.all_cpu.steal -stat_last.all_cpu.steal )*1000.0)/diff;

  /* copy last state */
  stat_last.pgpin  =stat_new.pgpin  ;
  stat_last.pgpout =stat_new.pgpout ;
  stat_last.swpin  =stat_new.swpin  ;
  stat_last.swpout =stat_new.swpout ;
  stat_last.context=stat_new.context;
  stat_last.forks  =stat_new.forks  ;
  stat_last.all_cpu.user  =stat_new.all_cpu.user  ;
  stat_last.all_cpu.nice  =stat_new.all_cpu.nice  ;
  stat_last.all_cpu.system=stat_new.all_cpu.system;
  stat_last.all_cpu.idle  =stat_new.all_cpu.idle  ;
  stat_last.all_cpu.system_real=stat_new.all_cpu.system_real;
  stat_last.all_cpu.irq   =stat_new.all_cpu.irq   ;
  stat_last.all_cpu.softirq=stat_new.all_cpu.softirq;
  stat_last.all_cpu.iowait=stat_new.all_cpu.iowait;
  stat_last.all_cpu.steal =stat_new.all_cpu.steal ;

/*
  if(debug){fprint(stderr,"sum=%lg\n",stat_cur.all_cpu.user+stat_cur.all_cpu.nice+stat_cur.all_cpu.system+
                  stat_cur.all_cpu.idle+stat_cur.all_cpu.irq+
                  stat_cur.all_cpu.softirq+stat_cur.all_cpu.iowait+stat_cur.all_cpu.steal);
  }
*/
  /* count increments and copy last results for each cpu */
  for(i=0;i<num_cpu;++i){
    stat_cur.cpu[i].user  =((stat_new.cpu[i].user  -stat_last.cpu[i].user  )*1000.0)/diff;
    stat_cur.cpu[i].nice  =((stat_new.cpu[i].nice  -stat_last.cpu[i].nice  )*1000.0)/diff;
    stat_cur.cpu[i].system=((stat_new.cpu[i].system-stat_last.cpu[i].system)*1000.0)/diff;
    stat_cur.cpu[i].idle  =((stat_new.cpu[i].idle  -stat_last.cpu[i].idle  )*1000.0)/diff;
    stat_cur.cpu[i].system_real=((stat_new.cpu[i].system_real-stat_last.cpu[i].system_real)*1000.0)/diff;
    stat_cur.cpu[i].irq   =((stat_new.cpu[i].irq   -stat_last.cpu[i].irq   )*1000.0)/diff;
    stat_cur.cpu[i].softirq=((stat_new.cpu[i].softirq-stat_last.cpu[i].softirq)*1000.0)/diff;
    stat_cur.cpu[i].iowait=((stat_new.cpu[i].iowait-stat_last.cpu[i].iowait)*1000.0)/diff;
    stat_cur.cpu[i].steal =((stat_new.cpu[i].steal -stat_last.cpu[i].steal )*1000.0)/diff;

    stat_last.cpu[i].user  =stat_new.cpu[i].user  ;
    stat_last.cpu[i].nice  =stat_new.cpu[i].nice  ;
    stat_last.cpu[i].system=stat_new.cpu[i].system;
    stat_last.cpu[i].idle  =stat_new.cpu[i].idle  ;
    stat_last.cpu[i].system_real=stat_new.cpu[i].system_real;
    stat_last.cpu[i].irq   =stat_new.cpu[i].irq   ;
    stat_last.cpu[i].softirq=stat_new.cpu[i].softirq;
    stat_last.cpu[i].iowait=stat_new.cpu[i].iowait;
    stat_last.cpu[i].steal =stat_new.cpu[i].steal ;
  }

  lasttime.tv_sec =curtime.tv_sec;
  lasttime.tv_usec=curtime.tv_usec;
  return 0;
}


/* --------------------------------------------
     reads loadaverages in 1, 5 and 15 miniutes into
     loadavg_1 loadavg_5 loadavg_15

     returns 1 if error
             0 if ok
   -------------------------------------------- */
int get_info_loadavg(void){
  char *buf;

  if(lseek(loadavg_fd,0,SEEK_SET)!=0){
    /* reopen file */
    close(loadavg_fd);
    if((loadavg_fd=open("/proc/loadavg",O_RDONLY))<0){
      print("*Cannot open /proc/loadavg.");
      return 1;
    }
  }

  if((buf=getstr(loadavg_fd))!=NULL){
    sscanf(buf,"%lg %lg %lg",&loadavg1_cur,&loadavg5_cur,&loadavg15_cur);
  }
  else{
    print("*cannot read /proc/loadavg");
    return 1;
  }
  return 0;
}


/* --------------------------------------------
     reads memory statistics into
           stat_curr

     returns 1 if error
             0 if ok
   -------------------------------------------- */
int get_info_mem(void){
  char *buf;

  if(lseek(mem_fd,0,SEEK_SET)!=0){
    /* reopen file */
    close(mem_fd);
    if((loadavg_fd=open("/proc/meminfo",O_RDONLY))<0){
      print("*Cannot open /proc/meminfo.");
      return 1;
    }
  }

  while((buf=getstr(mem_fd))!=NULL){
    if(strncmp(buf,"MemTotal",8)==0){
      sscanf(buf+10,"%lg",&stat_cur.mem_total);
    }
    else if(strncmp(buf,"MemFree",7)==0){
      sscanf(buf+9,"%lg",&stat_cur.mem_free);
    }
    else if(strncmp(buf,"MemShared",9)==0){
      sscanf(buf+11,"%lg",&stat_cur.mem_shared);
    }
    else if(strncmp(buf,"Buffers",7)==0){
      sscanf(buf+9,"%lg",&stat_cur.mem_buff);
    }
    else if(strncmp(buf,"Cached",6)==0){
      sscanf(buf+8,"%lg",&stat_cur.mem_cached);
    }
    else if(strncmp(buf,"SwapTotal",9)==0){
      sscanf(buf+11,"%lg",&stat_cur.mem_swp);
    }
    else if(strncmp(buf,"SwapFree",8)==0){
      sscanf(buf+10,"%lg",&stat_cur.mem_swp_free);
    }
  }
  return 0;
}


/* --------------------------------------------
     reads network statistics into
           net_curr[]

     returns 1 if error
             0 if ok
   -------------------------------------------- */
int get_info_net(void){
  char *buf;
  struct netstat netstat_tmp;
  static struct timeval curtime,lasttime={0,0};
  int i,j;

  if(lseek(net_fd,0,SEEK_SET)!=0){
    /* reopen file */
    close(net_fd);
    if((net_fd=open("/proc/net/dev",O_RDONLY))<0){
      print("*Cannot open /proc/net/dev.");
      return 1;
    }
  }

  /* skip legend */
  getstr(net_fd);
  getstr(net_fd);

  /* read info */
  while((buf=getstr(net_fd))!=NULL){
    /* get iface name */
    /* skip spaces */
    i=0;
    while(buf[i]==' ' && buf[i]!='\0')
      ++i;
    j=i;
    while(buf[j]!=':' && buf[j]!='\0')
      ++j;
    buf[j++]='\0';
    strcpy(netstat_tmp.if_name,buf+i);
    sscanf(buf+j," %lg %lg %lg %lg %lg %lg %lg %lg %lg %lg %lg %lg %lg %lg %lg %lg",
           &netstat_tmp.rcv_bytes,
           &netstat_tmp.rcv_pckts,
           &netstat_tmp.rcv_errs,
           &netstat_tmp.rcv_drop,
           &netstat_tmp.rcv_fifo,
           &netstat_tmp.rcv_frame,
           &netstat_tmp.rcv_compr,
           &netstat_tmp.rcv_multi,
           &netstat_tmp.sent_bytes,
           &netstat_tmp.sent_pckts,
           &netstat_tmp.sent_errs,
           &netstat_tmp.sent_drop,
           &netstat_tmp.sent_fifo,
           &netstat_tmp.sent_colls,
           &netstat_tmp.sent_carr,
           &netstat_tmp.sent_compr);
    /* find match for interface */
    for(i=0;i<num_if;++i){
      if(strcmp(netstat_tmp.if_name,netstat_cur[i].if_name)==0){

        /* copy values */
        netstat_new[i].rcv_bytes  = netstat_tmp.rcv_bytes;
        netstat_new[i].rcv_pckts  = netstat_tmp.rcv_pckts;
        netstat_new[i].rcv_errs   = netstat_tmp.rcv_errs;
        netstat_new[i].rcv_drop   = netstat_tmp.rcv_drop;
        netstat_new[i].rcv_fifo   = netstat_tmp.rcv_fifo;
        netstat_new[i].rcv_frame  = netstat_tmp.rcv_frame;
        netstat_new[i].rcv_compr  = netstat_tmp.rcv_compr;
        netstat_new[i].rcv_multi  = netstat_tmp.rcv_multi;
        netstat_new[i].sent_bytes = netstat_tmp.sent_bytes;
        netstat_new[i].sent_pckts = netstat_tmp.sent_pckts;
        netstat_new[i].sent_errs  = netstat_tmp.sent_errs;
        netstat_new[i].sent_drop  = netstat_tmp.sent_drop;
        netstat_new[i].sent_fifo  = netstat_tmp.sent_fifo;
        netstat_new[i].sent_colls = netstat_tmp.sent_colls;
        netstat_new[i].sent_carr  = netstat_tmp.sent_carr;
        netstat_new[i].sent_compr = netstat_tmp.sent_compr;
        break;
      }
    }
  }

  gettimeofday(&curtime,NULL);

  /* count time interval in milliseconds */
  diff=(curtime.tv_sec-lasttime.tv_sec)*1000+(curtime.tv_usec-lasttime.tv_usec)/1000;

  if(diff==0)
    diff=1000;

  /* count increments and copy last results for each cpu */
  for(i=0;i<num_if;++i){
    netstat_cur[i].rcv_bytes  = ((netstat_new[i].rcv_bytes  - netstat_last[i].rcv_bytes)*1000)/diff;
    netstat_cur[i].rcv_pckts  = ((netstat_new[i].rcv_pckts  - netstat_last[i].rcv_pckts)*1000)/diff;
    netstat_cur[i].rcv_errs   = ((netstat_new[i].rcv_errs   - netstat_last[i].rcv_errs)*1000)/diff;
    netstat_cur[i].rcv_drop   = ((netstat_new[i].rcv_drop   - netstat_last[i].rcv_drop)*1000)/diff;
    netstat_cur[i].rcv_fifo   = ((netstat_new[i].rcv_fifo   - netstat_last[i].rcv_fifo)*1000)/diff;
    netstat_cur[i].rcv_frame  = ((netstat_new[i].rcv_frame  - netstat_last[i].rcv_frame)*1000)/diff;
    netstat_cur[i].rcv_compr  = ((netstat_new[i].rcv_compr  - netstat_last[i].rcv_compr)*1000)/diff;
    netstat_cur[i].rcv_multi  = ((netstat_new[i].rcv_multi  - netstat_last[i].rcv_multi)*1000)/diff;
    netstat_cur[i].sent_bytes = ((netstat_new[i].sent_bytes - netstat_last[i].sent_bytes)*1000)/diff;
    netstat_cur[i].sent_pckts = ((netstat_new[i].sent_pckts - netstat_last[i].sent_pckts)*1000)/diff;
    netstat_cur[i].sent_errs  = ((netstat_new[i].sent_errs  - netstat_last[i].sent_errs)*1000)/diff;
    netstat_cur[i].sent_drop  = ((netstat_new[i].sent_drop  - netstat_last[i].sent_drop)*1000)/diff;
    netstat_cur[i].sent_fifo  = ((netstat_new[i].sent_fifo  - netstat_last[i].sent_fifo)*1000)/diff;
    netstat_cur[i].sent_colls = ((netstat_new[i].sent_colls - netstat_last[i].sent_colls)*1000)/diff;
    netstat_cur[i].sent_carr  = ((netstat_new[i].sent_carr  - netstat_last[i].sent_carr)*1000)/diff;
    netstat_cur[i].sent_compr = ((netstat_new[i].sent_compr - netstat_last[i].sent_compr)*1000)/diff;

    /* copy data */
    netstat_last[i].rcv_bytes   = netstat_new[i].rcv_bytes;
    netstat_last[i].rcv_pckts   = netstat_new[i].rcv_pckts;
    netstat_last[i].rcv_errs    = netstat_new[i].rcv_errs;
    netstat_last[i].rcv_drop    = netstat_new[i].rcv_drop;
    netstat_last[i].rcv_fifo    = netstat_new[i].rcv_fifo;
    netstat_last[i].rcv_frame   = netstat_new[i].rcv_frame;
    netstat_last[i].rcv_compr   = netstat_new[i].rcv_compr;
    netstat_last[i].rcv_multi   = netstat_new[i].rcv_multi;
    netstat_last[i].sent_bytes  = netstat_new[i].sent_bytes;
    netstat_last[i].sent_pckts  = netstat_new[i].sent_pckts;
    netstat_last[i].sent_errs   = netstat_new[i].sent_errs;
    netstat_last[i].sent_drop   = netstat_new[i].sent_drop;
    netstat_last[i].sent_fifo   = netstat_new[i].sent_fifo;
    netstat_last[i].sent_colls  = netstat_new[i].sent_colls;
    netstat_last[i].sent_carr   = netstat_new[i].sent_carr;
    netstat_last[i].sent_compr  = netstat_new[i].sent_compr;
  }

  lasttime.tv_sec =curtime.tv_sec;
  lasttime.tv_usec=curtime.tv_usec;
  return 0;
}

/* --------------------------------------------
     reads nfs client statistics into
           nfs_curr

     returns 1 if error
             0 if ok
   -------------------------------------------- */
int get_info_nfs(void){
  char *buf;
  static struct timeval curtime,lasttime={0,0};

  if(lseek(nfs_fd,0,SEEK_SET)!=0){
    /* reopen file */
    close(nfs_fd);
    if((nfs_fd=open("/proc/net/rpc/nfs",O_RDONLY))<0){
      print("*Cannot open /proc/net/rpc/nfs.");
      return 1;
    }
  }

  /* read info */
  while((buf=getstr(nfs_fd))!=NULL){
    if(strncmp(buf,"rpc",3)==0){
      sscanf(buf+4,"%lg %lg %lg",
             &nfsstat_new.nfs_count,
             &nfsstat_new.nfs_retrans,
             &nfsstat_new.nfs_auth);
      continue;
    }
    if(strncmp(buf,"proc2",5)==0){
      sscanf(buf+6,"%lg %lg %lg %lg %lg %lg %lg %lg %lg %lg %lg %lg %lg %lg %lg %lg %lg %lg",
             &nfsstat_new.nfs2_null,
             &nfsstat_new.nfs2_getattr,
             &nfsstat_new.nfs2_setattr,
             &nfsstat_new.nfs2_root,
             &nfsstat_new.nfs2_lookup,
             &nfsstat_new.nfs2_readlink,
             &nfsstat_new.nfs2_read,
             &nfsstat_new.nfs2_writechache,
             &nfsstat_new.nfs2_write,
             &nfsstat_new.nfs2_create,
             &nfsstat_new.nfs2_remove,
             &nfsstat_new.nfs2_rename,
             &nfsstat_new.nfs2_link,
             &nfsstat_new.nfs2_symlink,
             &nfsstat_new.nfs2_mkdir,
             &nfsstat_new.nfs2_rmdir,
             &nfsstat_new.nfs2_readdir,
             &nfsstat_new.nfs2_statfs
             );
      continue;
    }
    if(strncmp(buf,"proc3",5)==0){
      sscanf(buf+6,"%lg %lg %lg %lg %lg %lg %lg %lg %lg %lg %lg %lg %lg %lg %lg %lg %lg %lg %lg %lg %lg %lg",
             &nfsstat_new.nfs3_null,
             &nfsstat_new.nfs3_getattr,
             &nfsstat_new.nfs3_setattr,
             &nfsstat_new.nfs3_lookup,
             &nfsstat_new.nfs3_access,
             &nfsstat_new.nfs3_readlink,
             &nfsstat_new.nfs3_read,
             &nfsstat_new.nfs3_write,
             &nfsstat_new.nfs3_create,
             &nfsstat_new.nfs3_mkdir,
             &nfsstat_new.nfs3_symlink,
             &nfsstat_new.nfs3_mknod,
             &nfsstat_new.nfs3_remove,
             &nfsstat_new.nfs3_rmdir,
             &nfsstat_new.nfs3_rename,
             &nfsstat_new.nfs3_link,
             &nfsstat_new.nfs3_readdir,
             &nfsstat_new.nfs3_readdirplus,
             &nfsstat_new.nfs3_fsstat,
             &nfsstat_new.nfs3_fsinfo,
             &nfsstat_new.nfs3_pathconf,
             &nfsstat_new.nfs3_commit
             );
      continue;
    }
  }

  gettimeofday(&curtime,NULL);

  /* count time interval in milliseconds */
  diff=(curtime.tv_sec-lasttime.tv_sec)*1000+(curtime.tv_usec-lasttime.tv_usec)/1000;

  if(diff==0)
    diff=1000;

  nfsstat_cur.nfs_count        = ((nfsstat_new.nfs_count       - nfsstat_last.nfs_count       )*1000)/diff;
  nfsstat_cur.nfs_retrans      = ((nfsstat_new.nfs_retrans     - nfsstat_last.nfs_retrans     )*1000)/diff;
  nfsstat_cur.nfs_auth         = ((nfsstat_new.nfs_auth        - nfsstat_last.nfs_auth        )*1000)/diff;
  nfsstat_cur.nfs2_null        = ((nfsstat_new.nfs2_null       - nfsstat_last.nfs2_null       )*1000)/diff;
  nfsstat_cur.nfs2_getattr     = ((nfsstat_new.nfs2_getattr    - nfsstat_last.nfs2_getattr    )*1000)/diff;
  nfsstat_cur.nfs2_setattr     = ((nfsstat_new.nfs2_setattr    - nfsstat_last.nfs2_setattr    )*1000)/diff;
  nfsstat_cur.nfs2_root        = ((nfsstat_new.nfs2_root       - nfsstat_last.nfs2_root       )*1000)/diff;
  nfsstat_cur.nfs2_lookup      = ((nfsstat_new.nfs2_lookup     - nfsstat_last.nfs2_lookup     )*1000)/diff;
  nfsstat_cur.nfs2_readlink    = ((nfsstat_new.nfs2_readlink   - nfsstat_last.nfs2_readlink   )*1000)/diff;
  nfsstat_cur.nfs2_read        = ((nfsstat_new.nfs2_read       - nfsstat_last.nfs2_read       )*1000)/diff;
  nfsstat_cur.nfs2_writechache = ((nfsstat_new.nfs2_writechache- nfsstat_last.nfs2_writechache)*1000)/diff;
  nfsstat_cur.nfs2_write       = ((nfsstat_new.nfs2_write      - nfsstat_last.nfs2_write      )*1000)/diff;
  nfsstat_cur.nfs2_create      = ((nfsstat_new.nfs2_create     - nfsstat_last.nfs2_create     )*1000)/diff;
  nfsstat_cur.nfs2_remove      = ((nfsstat_new.nfs2_remove     - nfsstat_last.nfs2_remove     )*1000)/diff;
  nfsstat_cur.nfs2_rename      = ((nfsstat_new.nfs2_rename     - nfsstat_last.nfs2_rename     )*1000)/diff;
  nfsstat_cur.nfs2_link        = ((nfsstat_new.nfs2_link       - nfsstat_last.nfs2_link       )*1000)/diff;
  nfsstat_cur.nfs2_symlink     = ((nfsstat_new.nfs2_symlink    - nfsstat_last.nfs2_symlink    )*1000)/diff;
  nfsstat_cur.nfs2_mkdir       = ((nfsstat_new.nfs2_mkdir      - nfsstat_last.nfs2_mkdir      )*1000)/diff;
  nfsstat_cur.nfs2_rmdir       = ((nfsstat_new.nfs2_rmdir      - nfsstat_last.nfs2_rmdir      )*1000)/diff;
  nfsstat_cur.nfs2_readdir     = ((nfsstat_new.nfs2_readdir    - nfsstat_last.nfs2_readdir    )*1000)/diff;
  nfsstat_cur.nfs2_statfs      = ((nfsstat_new.nfs2_statfs     - nfsstat_last.nfs2_statfs     )*1000)/diff;
  nfsstat_cur.nfs3_null        = ((nfsstat_new.nfs3_null       - nfsstat_last.nfs3_null       )*1000)/diff;
  nfsstat_cur.nfs3_getattr     = ((nfsstat_new.nfs3_getattr    - nfsstat_last.nfs3_getattr    )*1000)/diff;
  nfsstat_cur.nfs3_setattr     = ((nfsstat_new.nfs3_setattr    - nfsstat_last.nfs3_setattr    )*1000)/diff;
  nfsstat_cur.nfs3_lookup      = ((nfsstat_new.nfs3_lookup     - nfsstat_last.nfs3_lookup     )*1000)/diff;
  nfsstat_cur.nfs3_access      = ((nfsstat_new.nfs3_access     - nfsstat_last.nfs3_access     )*1000)/diff;
  nfsstat_cur.nfs3_readlink    = ((nfsstat_new.nfs3_readlink   - nfsstat_last.nfs3_readlink   )*1000)/diff;
  nfsstat_cur.nfs3_read        = ((nfsstat_new.nfs3_read       - nfsstat_last.nfs3_read       )*1000)/diff;
  nfsstat_cur.nfs3_write       = ((nfsstat_new.nfs3_write      - nfsstat_last.nfs3_write      )*1000)/diff;
  nfsstat_cur.nfs3_create      = ((nfsstat_new.nfs3_create     - nfsstat_last.nfs3_create     )*1000)/diff;
  nfsstat_cur.nfs3_mkdir       = ((nfsstat_new.nfs3_mkdir      - nfsstat_last.nfs3_mkdir      )*1000)/diff;
  nfsstat_cur.nfs3_symlink     = ((nfsstat_new.nfs3_symlink    - nfsstat_last.nfs3_symlink    )*1000)/diff;
  nfsstat_cur.nfs3_mknod       = ((nfsstat_new.nfs3_mknod      - nfsstat_last.nfs3_mknod      )*1000)/diff;
  nfsstat_cur.nfs3_remove      = ((nfsstat_new.nfs3_remove     - nfsstat_last.nfs3_remove     )*1000)/diff;
  nfsstat_cur.nfs3_rmdir       = ((nfsstat_new.nfs3_rmdir      - nfsstat_last.nfs3_rmdir      )*1000)/diff;
  nfsstat_cur.nfs3_rename      = ((nfsstat_new.nfs3_rename     - nfsstat_last.nfs3_rename     )*1000)/diff;
  nfsstat_cur.nfs3_link        = ((nfsstat_new.nfs3_link       - nfsstat_last.nfs3_link       )*1000)/diff;
  nfsstat_cur.nfs3_readdir     = ((nfsstat_new.nfs3_readdir    - nfsstat_last.nfs3_readdir    )*1000)/diff;
  nfsstat_cur.nfs3_readdirplus = ((nfsstat_new.nfs3_readdirplus- nfsstat_last.nfs3_readdirplus)*1000)/diff;
  nfsstat_cur.nfs3_fsstat      = ((nfsstat_new.nfs3_fsstat     - nfsstat_last.nfs3_fsstat     )*1000)/diff;
  nfsstat_cur.nfs3_fsinfo      = ((nfsstat_new.nfs3_fsinfo     - nfsstat_last.nfs3_fsinfo     )*1000)/diff;
  nfsstat_cur.nfs3_pathconf    = ((nfsstat_new.nfs3_pathconf   - nfsstat_last.nfs3_pathconf   )*1000)/diff;
  nfsstat_cur.nfs3_commit      = ((nfsstat_new.nfs3_commit     - nfsstat_last.nfs3_commit     )*1000)/diff;

  /* copy data */
  nfsstat_last.nfs_count       = nfsstat_new.nfs_count;
  nfsstat_last.nfs_retrans     = nfsstat_new.nfs_retrans;
  nfsstat_last.nfs_auth        = nfsstat_new.nfs_auth;
  nfsstat_last.nfs2_null       = nfsstat_new.nfs2_null;
  nfsstat_last.nfs2_getattr    = nfsstat_new.nfs2_getattr;
  nfsstat_last.nfs2_setattr    = nfsstat_new.nfs2_setattr;
  nfsstat_last.nfs2_root       = nfsstat_new.nfs2_root;
  nfsstat_last.nfs2_lookup     = nfsstat_new.nfs2_lookup;
  nfsstat_last.nfs2_readlink   = nfsstat_new.nfs2_readlink;
  nfsstat_last.nfs2_read       = nfsstat_new.nfs2_read;
  nfsstat_last.nfs2_writechache= nfsstat_new.nfs2_writechache;
  nfsstat_last.nfs2_write      = nfsstat_new.nfs2_write;
  nfsstat_last.nfs2_create     = nfsstat_new.nfs2_create;
  nfsstat_last.nfs2_remove     = nfsstat_new.nfs2_remove;
  nfsstat_last.nfs2_rename     = nfsstat_new.nfs2_rename;
  nfsstat_last.nfs2_link       = nfsstat_new.nfs2_link;
  nfsstat_last.nfs2_symlink    = nfsstat_new.nfs2_symlink;
  nfsstat_last.nfs2_mkdir      = nfsstat_new.nfs2_mkdir;
  nfsstat_last.nfs2_rmdir      = nfsstat_new.nfs2_rmdir;
  nfsstat_last.nfs2_readdir    = nfsstat_new.nfs2_readdir;
  nfsstat_last.nfs2_statfs     = nfsstat_new.nfs2_statfs;
  nfsstat_last.nfs3_null       = nfsstat_new.nfs3_null;
  nfsstat_last.nfs3_getattr    = nfsstat_new.nfs3_getattr;
  nfsstat_last.nfs3_setattr    = nfsstat_new.nfs3_setattr;
  nfsstat_last.nfs3_lookup     = nfsstat_new.nfs3_lookup;
  nfsstat_last.nfs3_access     = nfsstat_new.nfs3_access;
  nfsstat_last.nfs3_readlink   = nfsstat_new.nfs3_readlink;
  nfsstat_last.nfs3_read       = nfsstat_new.nfs3_read;
  nfsstat_last.nfs3_write      = nfsstat_new.nfs3_write;
  nfsstat_last.nfs3_create     = nfsstat_new.nfs3_create;
  nfsstat_last.nfs3_mkdir      = nfsstat_new.nfs3_mkdir;
  nfsstat_last.nfs3_symlink    = nfsstat_new.nfs3_symlink;
  nfsstat_last.nfs3_mknod      = nfsstat_new.nfs3_mknod;
  nfsstat_last.nfs3_remove     = nfsstat_new.nfs3_remove;
  nfsstat_last.nfs3_rmdir      = nfsstat_new.nfs3_rmdir;
  nfsstat_last.nfs3_rename     = nfsstat_new.nfs3_rename;
  nfsstat_last.nfs3_link       = nfsstat_new.nfs3_link;
  nfsstat_last.nfs3_readdir    = nfsstat_new.nfs3_readdir;
  nfsstat_last.nfs3_readdirplus= nfsstat_new.nfs3_readdirplus;
  nfsstat_last.nfs3_fsstat     = nfsstat_new.nfs3_fsstat;
  nfsstat_last.nfs3_fsinfo     = nfsstat_new.nfs3_fsinfo;
  nfsstat_last.nfs3_pathconf   = nfsstat_new.nfs3_pathconf;
  nfsstat_last.nfs3_commit     = nfsstat_new.nfs3_commit;

  lasttime.tv_sec =curtime.tv_sec;
  lasttime.tv_usec=curtime.tv_usec;
  return 0;
}

/* --------------------------------------------
     reads nfs daemon statistics into
           nfsd_curr

     returns 1 if error
             0 if ok
   -------------------------------------------- */
int get_info_nfsd(void){
  char *buf;
  static struct timeval curtime,lasttime={0,0};

  if(lseek(nfsd_fd,0,SEEK_SET)!=0){
    /* reopen file */
    close(nfsd_fd);
    if((nfsd_fd=open("/proc/net/rpc/nfsd",O_RDONLY))<0){
      print("*Cannot open /proc/net/rpc/nfsd.");
      return 1;
    }
  }

  /* read info */
  while((buf=getstr(nfsd_fd))!=NULL){
    if(strncmp(buf,"rpc",3)==0){
      sscanf(buf+4,"%lg %lg %lg %lg %lg",
             &nfsdstat_new.nfsd_calls,
             &nfsdstat_new.nfsd_badcalls,
             &nfsdstat_new.nfsd_badauth,
             &nfsdstat_new.nfsd_badclnt,
             &nfsdstat_new.nfsd_xdrcall);
      continue;
    }
    if(strncmp(buf,"proc2",5)==0){
      sscanf(buf+6,"%lg %lg %lg %lg %lg %lg %lg %lg %lg %lg %lg %lg %lg %lg %lg %lg %lg %lg",
             &nfsdstat_new.nfsd2_null,
             &nfsdstat_new.nfsd2_getattr,
             &nfsdstat_new.nfsd2_setattr,
             &nfsdstat_new.nfsd2_root,
             &nfsdstat_new.nfsd2_lookup,
             &nfsdstat_new.nfsd2_readlink,
             &nfsdstat_new.nfsd2_read,
             &nfsdstat_new.nfsd2_writechache,
             &nfsdstat_new.nfsd2_write,
             &nfsdstat_new.nfsd2_create,
             &nfsdstat_new.nfsd2_remove,
             &nfsdstat_new.nfsd2_rename,
             &nfsdstat_new.nfsd2_link,
             &nfsdstat_new.nfsd2_symlink,
             &nfsdstat_new.nfsd2_mkdir,
             &nfsdstat_new.nfsd2_rmdir,
             &nfsdstat_new.nfsd2_readdir,
             &nfsdstat_new.nfsd2_statfs
             );
      continue;
    }
    if(strncmp(buf,"proc3",5)==0){
      sscanf(buf+6,"%lg %lg %lg %lg %lg %lg %lg %lg %lg %lg %lg %lg %lg %lg %lg %lg %lg %lg %lg %lg %lg %lg",
             &nfsdstat_new.nfsd3_null,
             &nfsdstat_new.nfsd3_getattr,
             &nfsdstat_new.nfsd3_setattr,
             &nfsdstat_new.nfsd3_lookup,
             &nfsdstat_new.nfsd3_access,
             &nfsdstat_new.nfsd3_readlink,
             &nfsdstat_new.nfsd3_read,
             &nfsdstat_new.nfsd3_write,
             &nfsdstat_new.nfsd3_create,
             &nfsdstat_new.nfsd3_mkdir,
             &nfsdstat_new.nfsd3_symlink,
             &nfsdstat_new.nfsd3_mknod,
             &nfsdstat_new.nfsd3_remove,
             &nfsdstat_new.nfsd3_rmdir,
             &nfsdstat_new.nfsd3_rename,
             &nfsdstat_new.nfsd3_link,
             &nfsdstat_new.nfsd3_readdir,
             &nfsdstat_new.nfsd3_readdirplus,
             &nfsdstat_new.nfsd3_fsstat,
             &nfsdstat_new.nfsd3_fsinfo,
             &nfsdstat_new.nfsd3_pathconf,
             &nfsdstat_new.nfsd3_commit
             );
      continue;
    }
    if(strncmp(buf,"rc ",3)==0){
      sscanf(buf+3,"%lg %lg %lg",
             &nfsdstat_new.nfsd3_cache,
             &nfsdstat_new.nfsd3_miss,
             &nfsdstat_new.nfsd3_nocache
             );
      continue;
    }
    if(strncmp(buf,"io ",3)==0){
      sscanf(buf+3,"%lg %lg",
             &nfsdstat_new.nfsd3_io_read,
             &nfsdstat_new.nfsd3_io_write
             );
      continue;
    }
    if(strncmp(buf,"net ",4)==0){
      sscanf(buf+4,"%lg",
             &nfsdstat_new.nfsd3_netcount
             );
      continue;
    }
    if(strncmp(buf,"th ",3)==0){
      sscanf(buf+3,"%lg %lg %lg %lg %lg %lg %lg %lg %lg %lg %lg",
             &nfsdstat_new.nfsd3_thread_sec,
             &nfsdstat_new.nfsd3_thread_10,
             &nfsdstat_new.nfsd3_thread_20,
             &nfsdstat_new.nfsd3_thread_30,
             &nfsdstat_new.nfsd3_thread_40,
             &nfsdstat_new.nfsd3_thread_50,
             &nfsdstat_new.nfsd3_thread_60,
             &nfsdstat_new.nfsd3_thread_70,
             &nfsdstat_new.nfsd3_thread_80,
             &nfsdstat_new.nfsd3_thread_90,
             &nfsdstat_new.nfsd3_thread_100
             );
      continue;
    }
  }

  gettimeofday(&curtime,NULL);

  /* count time interval in milliseconds */
  diff=(curtime.tv_sec-lasttime.tv_sec)*1000+(curtime.tv_usec-lasttime.tv_usec)/1000;

  if(diff==0)
    diff=1000;

  nfsdstat_cur.nfsd_calls        = ((nfsdstat_new.nfsd_calls       - nfsdstat_last.nfsd_calls       )*1000)/diff;
  nfsdstat_cur.nfsd_badcalls     = ((nfsdstat_new.nfsd_badcalls    - nfsdstat_last.nfsd_badcalls    )*1000)/diff;
  nfsdstat_cur.nfsd_badauth      = ((nfsdstat_new.nfsd_badauth     - nfsdstat_last.nfsd_badauth     )*1000)/diff;
  nfsdstat_cur.nfsd_badclnt      = ((nfsdstat_new.nfsd_badclnt     - nfsdstat_last.nfsd_badclnt     )*1000)/diff;
  nfsdstat_cur.nfsd_xdrcall      = ((nfsdstat_new.nfsd_xdrcall     - nfsdstat_last.nfsd_xdrcall     )*1000)/diff;
  nfsdstat_cur.nfsd2_null        = ((nfsdstat_new.nfsd2_null       - nfsdstat_last.nfsd2_null       )*1000)/diff;
  nfsdstat_cur.nfsd2_getattr     = ((nfsdstat_new.nfsd2_getattr    - nfsdstat_last.nfsd2_getattr    )*1000)/diff;
  nfsdstat_cur.nfsd2_setattr     = ((nfsdstat_new.nfsd2_setattr    - nfsdstat_last.nfsd2_setattr    )*1000)/diff;
  nfsdstat_cur.nfsd2_root        = ((nfsdstat_new.nfsd2_root       - nfsdstat_last.nfsd2_root       )*1000)/diff;
  nfsdstat_cur.nfsd2_lookup      = ((nfsdstat_new.nfsd2_lookup     - nfsdstat_last.nfsd2_lookup     )*1000)/diff;
  nfsdstat_cur.nfsd2_readlink    = ((nfsdstat_new.nfsd2_readlink   - nfsdstat_last.nfsd2_readlink   )*1000)/diff;
  nfsdstat_cur.nfsd2_read        = ((nfsdstat_new.nfsd2_read       - nfsdstat_last.nfsd2_read       )*1000)/diff;
  nfsdstat_cur.nfsd2_writechache = ((nfsdstat_new.nfsd2_writechache- nfsdstat_last.nfsd2_writechache)*1000)/diff;
  nfsdstat_cur.nfsd2_write       = ((nfsdstat_new.nfsd2_write      - nfsdstat_last.nfsd2_write      )*1000)/diff;
  nfsdstat_cur.nfsd2_create      = ((nfsdstat_new.nfsd2_create     - nfsdstat_last.nfsd2_create     )*1000)/diff;
  nfsdstat_cur.nfsd2_remove      = ((nfsdstat_new.nfsd2_remove     - nfsdstat_last.nfsd2_remove     )*1000)/diff;
  nfsdstat_cur.nfsd2_rename      = ((nfsdstat_new.nfsd2_rename     - nfsdstat_last.nfsd2_rename     )*1000)/diff;
  nfsdstat_cur.nfsd2_link        = ((nfsdstat_new.nfsd2_link       - nfsdstat_last.nfsd2_link       )*1000)/diff;
  nfsdstat_cur.nfsd2_symlink     = ((nfsdstat_new.nfsd2_symlink    - nfsdstat_last.nfsd2_symlink    )*1000)/diff;
  nfsdstat_cur.nfsd2_mkdir       = ((nfsdstat_new.nfsd2_mkdir      - nfsdstat_last.nfsd2_mkdir      )*1000)/diff;
  nfsdstat_cur.nfsd2_rmdir       = ((nfsdstat_new.nfsd2_rmdir      - nfsdstat_last.nfsd2_rmdir      )*1000)/diff;
  nfsdstat_cur.nfsd2_readdir     = ((nfsdstat_new.nfsd2_readdir    - nfsdstat_last.nfsd2_readdir    )*1000)/diff;
  nfsdstat_cur.nfsd2_statfs      = ((nfsdstat_new.nfsd2_statfs     - nfsdstat_last.nfsd2_statfs     )*1000)/diff;
  nfsdstat_cur.nfsd3_null        = ((nfsdstat_new.nfsd3_null       - nfsdstat_last.nfsd3_null       )*1000)/diff;
  nfsdstat_cur.nfsd3_getattr     = ((nfsdstat_new.nfsd3_getattr    - nfsdstat_last.nfsd3_getattr    )*1000)/diff;
  nfsdstat_cur.nfsd3_setattr     = ((nfsdstat_new.nfsd3_setattr    - nfsdstat_last.nfsd3_setattr    )*1000)/diff;
  nfsdstat_cur.nfsd3_lookup      = ((nfsdstat_new.nfsd3_lookup     - nfsdstat_last.nfsd3_lookup     )*1000)/diff;
  nfsdstat_cur.nfsd3_access      = ((nfsdstat_new.nfsd3_access     - nfsdstat_last.nfsd3_access     )*1000)/diff;
  nfsdstat_cur.nfsd3_readlink    = ((nfsdstat_new.nfsd3_readlink   - nfsdstat_last.nfsd3_readlink   )*1000)/diff;
  nfsdstat_cur.nfsd3_read        = ((nfsdstat_new.nfsd3_read       - nfsdstat_last.nfsd3_read       )*1000)/diff;
  nfsdstat_cur.nfsd3_write       = ((nfsdstat_new.nfsd3_write      - nfsdstat_last.nfsd3_write      )*1000)/diff;
  nfsdstat_cur.nfsd3_create      = ((nfsdstat_new.nfsd3_create     - nfsdstat_last.nfsd3_create     )*1000)/diff;
  nfsdstat_cur.nfsd3_mkdir       = ((nfsdstat_new.nfsd3_mkdir      - nfsdstat_last.nfsd3_mkdir      )*1000)/diff;
  nfsdstat_cur.nfsd3_symlink     = ((nfsdstat_new.nfsd3_symlink    - nfsdstat_last.nfsd3_symlink    )*1000)/diff;
  nfsdstat_cur.nfsd3_mknod       = ((nfsdstat_new.nfsd3_mknod      - nfsdstat_last.nfsd3_mknod      )*1000)/diff;
  nfsdstat_cur.nfsd3_remove      = ((nfsdstat_new.nfsd3_remove     - nfsdstat_last.nfsd3_remove     )*1000)/diff;
  nfsdstat_cur.nfsd3_rmdir       = ((nfsdstat_new.nfsd3_rmdir      - nfsdstat_last.nfsd3_rmdir      )*1000)/diff;
  nfsdstat_cur.nfsd3_rename      = ((nfsdstat_new.nfsd3_rename     - nfsdstat_last.nfsd3_rename     )*1000)/diff;
  nfsdstat_cur.nfsd3_link        = ((nfsdstat_new.nfsd3_link       - nfsdstat_last.nfsd3_link       )*1000)/diff;
  nfsdstat_cur.nfsd3_readdir     = ((nfsdstat_new.nfsd3_readdir    - nfsdstat_last.nfsd3_readdir    )*1000)/diff;
  nfsdstat_cur.nfsd3_readdirplus = ((nfsdstat_new.nfsd3_readdirplus- nfsdstat_last.nfsd3_readdirplus)*1000)/diff;
  nfsdstat_cur.nfsd3_fsstat      = ((nfsdstat_new.nfsd3_fsstat     - nfsdstat_last.nfsd3_fsstat     )*1000)/diff;
  nfsdstat_cur.nfsd3_fsinfo      = ((nfsdstat_new.nfsd3_fsinfo     - nfsdstat_last.nfsd3_fsinfo     )*1000)/diff;
  nfsdstat_cur.nfsd3_pathconf    = ((nfsdstat_new.nfsd3_pathconf   - nfsdstat_last.nfsd3_pathconf   )*1000)/diff;
  nfsdstat_cur.nfsd3_commit      = ((nfsdstat_new.nfsd3_commit     - nfsdstat_last.nfsd3_commit     )*1000)/diff;

  nfsdstat_cur.nfsd3_cache       = ((nfsdstat_new.nfsd3_cache      - nfsdstat_last.nfsd3_cache      )*1000)/diff;
  nfsdstat_cur.nfsd3_miss        = ((nfsdstat_new.nfsd3_miss       - nfsdstat_last.nfsd3_miss       )*1000)/diff;
  nfsdstat_cur.nfsd3_nocache     = ((nfsdstat_new.nfsd3_nocache    - nfsdstat_last.nfsd3_nocache    )*1000)/diff;
  nfsdstat_cur.nfsd3_io_read     = ((nfsdstat_new.nfsd3_io_read    - nfsdstat_last.nfsd3_io_read    )*1000)/diff;
  nfsdstat_cur.nfsd3_io_write    = ((nfsdstat_new.nfsd3_io_write   - nfsdstat_last.nfsd3_io_write   )*1000)/diff;
  nfsdstat_cur.nfsd3_netcount    = ((nfsdstat_new.nfsd3_netcount   - nfsdstat_last.nfsd3_netcount   )*1000)/diff;
  nfsdstat_cur.nfsd3_thread_sec  = ((nfsdstat_new.nfsd3_thread_sec - nfsdstat_last.nfsd3_thread_sec )*1000)/diff;
  nfsdstat_cur.nfsd3_thread_10   = ((nfsdstat_new.nfsd3_thread_10  - nfsdstat_last.nfsd3_thread_10  )*1000)/diff;
  nfsdstat_cur.nfsd3_thread_20   = ((nfsdstat_new.nfsd3_thread_20  - nfsdstat_last.nfsd3_thread_20  )*1000)/diff;
  nfsdstat_cur.nfsd3_thread_30   = ((nfsdstat_new.nfsd3_thread_30  - nfsdstat_last.nfsd3_thread_30  )*1000)/diff;
  nfsdstat_cur.nfsd3_thread_40   = ((nfsdstat_new.nfsd3_thread_40  - nfsdstat_last.nfsd3_thread_40  )*1000)/diff;
  nfsdstat_cur.nfsd3_thread_50   = ((nfsdstat_new.nfsd3_thread_50  - nfsdstat_last.nfsd3_thread_50  )*1000)/diff;
  nfsdstat_cur.nfsd3_thread_60   = ((nfsdstat_new.nfsd3_thread_60  - nfsdstat_last.nfsd3_thread_60  )*1000)/diff;
  nfsdstat_cur.nfsd3_thread_70   = ((nfsdstat_new.nfsd3_thread_70  - nfsdstat_last.nfsd3_thread_70  )*1000)/diff;
  nfsdstat_cur.nfsd3_thread_80   = ((nfsdstat_new.nfsd3_thread_80  - nfsdstat_last.nfsd3_thread_80  )*1000)/diff;
  nfsdstat_cur.nfsd3_thread_90   = ((nfsdstat_new.nfsd3_thread_90  - nfsdstat_last.nfsd3_thread_90  )*1000)/diff;
  nfsdstat_cur.nfsd3_thread_100  = ((nfsdstat_new.nfsd3_thread_100 - nfsdstat_last.nfsd3_thread_100 )*1000)/diff;

  /* copy data */
  nfsdstat_last.nfsd_calls       = nfsdstat_new.nfsd_calls;
  nfsdstat_last.nfsd_badcalls    = nfsdstat_new.nfsd_badcalls;
  nfsdstat_last.nfsd_badauth     = nfsdstat_new.nfsd_badauth;
  nfsdstat_last.nfsd_badclnt     = nfsdstat_new.nfsd_badclnt;
  nfsdstat_last.nfsd_xdrcall     = nfsdstat_new.nfsd_xdrcall;
  nfsdstat_last.nfsd2_null       = nfsdstat_new.nfsd2_null;
  nfsdstat_last.nfsd2_getattr    = nfsdstat_new.nfsd2_getattr;
  nfsdstat_last.nfsd2_setattr    = nfsdstat_new.nfsd2_setattr;
  nfsdstat_last.nfsd2_root       = nfsdstat_new.nfsd2_root;
  nfsdstat_last.nfsd2_lookup     = nfsdstat_new.nfsd2_lookup;
  nfsdstat_last.nfsd2_readlink   = nfsdstat_new.nfsd2_readlink;
  nfsdstat_last.nfsd2_read       = nfsdstat_new.nfsd2_read;
  nfsdstat_last.nfsd2_writechache= nfsdstat_new.nfsd2_writechache;
  nfsdstat_last.nfsd2_write      = nfsdstat_new.nfsd2_write;
  nfsdstat_last.nfsd2_create     = nfsdstat_new.nfsd2_create;
  nfsdstat_last.nfsd2_remove     = nfsdstat_new.nfsd2_remove;
  nfsdstat_last.nfsd2_rename     = nfsdstat_new.nfsd2_rename;
  nfsdstat_last.nfsd2_link       = nfsdstat_new.nfsd2_link;
  nfsdstat_last.nfsd2_symlink    = nfsdstat_new.nfsd2_symlink;
  nfsdstat_last.nfsd2_mkdir      = nfsdstat_new.nfsd2_mkdir;
  nfsdstat_last.nfsd2_rmdir      = nfsdstat_new.nfsd2_rmdir;
  nfsdstat_last.nfsd2_readdir    = nfsdstat_new.nfsd2_readdir;
  nfsdstat_last.nfsd2_statfs     = nfsdstat_new.nfsd2_statfs;
  nfsdstat_last.nfsd3_null       = nfsdstat_new.nfsd3_null;
  nfsdstat_last.nfsd3_getattr    = nfsdstat_new.nfsd3_getattr;
  nfsdstat_last.nfsd3_setattr    = nfsdstat_new.nfsd3_setattr;
  nfsdstat_last.nfsd3_lookup     = nfsdstat_new.nfsd3_lookup;
  nfsdstat_last.nfsd3_access     = nfsdstat_new.nfsd3_access;
  nfsdstat_last.nfsd3_readlink   = nfsdstat_new.nfsd3_readlink;
  nfsdstat_last.nfsd3_read       = nfsdstat_new.nfsd3_read;
  nfsdstat_last.nfsd3_write      = nfsdstat_new.nfsd3_write;
  nfsdstat_last.nfsd3_create     = nfsdstat_new.nfsd3_create;
  nfsdstat_last.nfsd3_mkdir      = nfsdstat_new.nfsd3_mkdir;
  nfsdstat_last.nfsd3_symlink    = nfsdstat_new.nfsd3_symlink;
  nfsdstat_last.nfsd3_mknod      = nfsdstat_new.nfsd3_mknod;
  nfsdstat_last.nfsd3_remove     = nfsdstat_new.nfsd3_remove;
  nfsdstat_last.nfsd3_rmdir      = nfsdstat_new.nfsd3_rmdir;
  nfsdstat_last.nfsd3_rename     = nfsdstat_new.nfsd3_rename;
  nfsdstat_last.nfsd3_link       = nfsdstat_new.nfsd3_link;
  nfsdstat_last.nfsd3_readdir    = nfsdstat_new.nfsd3_readdir;
  nfsdstat_last.nfsd3_readdirplus= nfsdstat_new.nfsd3_readdirplus;
  nfsdstat_last.nfsd3_fsstat     = nfsdstat_new.nfsd3_fsstat;
  nfsdstat_last.nfsd3_fsinfo     = nfsdstat_new.nfsd3_fsinfo;
  nfsdstat_last.nfsd3_pathconf   = nfsdstat_new.nfsd3_pathconf;
  nfsdstat_last.nfsd3_commit     = nfsdstat_new.nfsd3_commit;

  nfsdstat_last.nfsd3_cache      = nfsdstat_new.nfsd3_cache;
  nfsdstat_last.nfsd3_miss       = nfsdstat_new.nfsd3_miss;
  nfsdstat_last.nfsd3_nocache    = nfsdstat_new.nfsd3_nocache;
  nfsdstat_last.nfsd3_io_read    = nfsdstat_new.nfsd3_io_read;
  nfsdstat_last.nfsd3_io_write   = nfsdstat_new.nfsd3_io_write;
  nfsdstat_last.nfsd3_netcount   = nfsdstat_new.nfsd3_netcount;
  nfsdstat_last.nfsd3_thread_sec = nfsdstat_new.nfsd3_thread_sec;
  nfsdstat_last.nfsd3_thread_10  = nfsdstat_new.nfsd3_thread_10;
  nfsdstat_last.nfsd3_thread_20  = nfsdstat_new.nfsd3_thread_20;
  nfsdstat_last.nfsd3_thread_30  = nfsdstat_new.nfsd3_thread_30;
  nfsdstat_last.nfsd3_thread_40  = nfsdstat_new.nfsd3_thread_40;
  nfsdstat_last.nfsd3_thread_50  = nfsdstat_new.nfsd3_thread_50;
  nfsdstat_last.nfsd3_thread_60  = nfsdstat_new.nfsd3_thread_60;
  nfsdstat_last.nfsd3_thread_70  = nfsdstat_new.nfsd3_thread_70;
  nfsdstat_last.nfsd3_thread_80  = nfsdstat_new.nfsd3_thread_80;
  nfsdstat_last.nfsd3_thread_90  = nfsdstat_new.nfsd3_thread_90;
  nfsdstat_last.nfsd3_thread_100 = nfsdstat_new.nfsd3_thread_100;

  lasttime.tv_sec =curtime.tv_sec;
  lasttime.tv_usec=curtime.tv_usec;
  return 0;
}

/* -------------------------------
   Test basic request

   Return - pointer to cpu_stat          if success,
          - pointer to stat_cur.all_cpu  if bad cpu,
          - NULL                         if no match
---------------------------------*/

inline struct cpu_stat * test_basic_stat(char *req, char *match){
  static int cpu;

  if(sscanf(req,match,&cpu)>0){
    if((cpu<num_cpu) && (cpu>=0)){
      return &(stat_cur.cpu[cpu]);
    }
    print("*no such cpu\n");
    return &stat_cur.all_cpu;
  }
  return NULL;
}
/* --------------------------
   process one request line
   print result
   -------------------------- */
void process_line(char *req){

  struct str_liptr *cur;
  struct cpu_stat  *basic_stat;
  int cpu;

/*  if(sscanf(req,"user%d",&cpu)>0){
    if((cpu<num_cpu) && (cpu>=0)){
      print("%f\n",stat_cur.cpu[cpu].user);
      return;
    }
    print("*no such cpu\n");
    return;
  }
*/
  if((basic_stat=test_basic_stat(req,"cpu_user%d"))!=NULL){
    if(basic_stat==&stat_cur.all_cpu)/* bad cpu */
      return;
    print("%.4lf\n",basic_stat->user);
    return;
  }

  if((basic_stat=test_basic_stat(req,"cpu_system%d"))!=NULL){
    if(basic_stat==&stat_cur.all_cpu)/* bad cpu */
      return;
    print("%.4lf\n",basic_stat->system);
    return;
  }

  if((basic_stat=test_basic_stat(req,"cpu_nice%d"))!=NULL){
    if(basic_stat==&stat_cur.all_cpu)/* bad cpu */
      return;
    print("%.4lf\n",basic_stat->nice);
    return;
  }

  if((basic_stat=test_basic_stat(req,"cpu_idle%d"))!=NULL){
    if(basic_stat==&stat_cur.all_cpu)/* bad cpu */
      return;
    print("%.4lf\n",basic_stat->idle);
    return;
  }

  if((basic_stat=test_basic_stat(req,"cpu_sys_real%d"))!=NULL){
    if(basic_stat==&stat_cur.all_cpu)/* bad cpu */
      return;
    print("%.4lf\n",basic_stat->system_real);
    return;
  }

  if((basic_stat=test_basic_stat(req,"cpu_irq%d"))!=NULL){
    if(basic_stat==&stat_cur.all_cpu)/* bad cpu */
      return;
    print("%.4lf\n",basic_stat->irq);
    return;
  }

  if((basic_stat=test_basic_stat(req,"cpu_softirq%d"))!=NULL){
    if(basic_stat==&stat_cur.all_cpu)/* bad cpu */
      return;
    print("%.4lf\n",basic_stat->softirq);
    return;
  }

  if((basic_stat=test_basic_stat(req,"cpu_iowait%d"))!=NULL){
    if(basic_stat==&stat_cur.all_cpu)/* bad cpu */
      return;
    print("%.4lf\n",basic_stat->iowait);
    return;
  }

  if((basic_stat=test_basic_stat(req,"cpu_steal%d"))!=NULL){
    if(basic_stat==&stat_cur.all_cpu)/* bad cpu */
      return;
    print("%.4lf\n",basic_stat->steal);
    return;
  }

  for(cur=&baselist[0];cur->str!=NULL;++cur){
    /* match ? */
    if(strcmp(req,cur->str)==0){
      print("%.4lf\n",*(cur->ptr));
      return;
    }
  }
  for(cur=&ifaceslist[0];cur->str!=NULL;++cur){
    /* match ? */
    if(strcmp(req,cur->str)==0){
      print("%.4lf\n",*(cur->ptr));
      return;
    }
  }
  print("*bad request\n");
}






/* main procedure */
int main(){

  char p,p2; /* for read */
  char *buf,request[MAX_REQ_LINE];
  int  i,j,start,req,flags;

  setlocale(LC_ALL,"C");

  if((stat26_fd=open("/proc/vmstat",O_RDONLY))>=0){
    k26=1;
  }
  stat_fd=open("/proc/stat",O_RDONLY);
  loadavg_fd=open("/proc/loadavg",O_RDONLY);
  mem_fd=open("/proc/meminfo",O_RDONLY);
  net_fd=open("/proc/net/dev",O_RDONLY);
  nfs_fd=open("/proc/net/rpc/nfs",O_RDONLY);
  nfsd_fd=open("/proc/net/rpc/nfsd",O_RDONLY);

  num_cpu=0;
  /* check out processors quantity */
  while((buf=getstr(stat_fd))!=NULL){
    sscanf(buf,"cpu%d",&num_cpu);
    if(sscanf(buf,"cpu %d %d %d %d %d %d %d %d",&i,&i,&i,&i,&i,&i,&i,&i)==8)
      stat_long=1;
  }
  ++num_cpu;
  stat_cur.cpu=(struct cpu_stat *)malloc(sizeof(struct cpu_stat)*num_cpu);
  stat_new.cpu=(struct cpu_stat *)malloc(sizeof(struct cpu_stat)*num_cpu);
  stat_last.cpu=(struct cpu_stat *)malloc(sizeof(struct cpu_stat)*num_cpu);

  /* get all ifaces names */
  /* skip legend */
  getstr(net_fd);
  getstr(net_fd);

  num_if=0;
  while(buf=getstr(net_fd)){
    ++num_if;
  }
  close(net_fd);

  netstat_cur=(struct netstat *)malloc(sizeof(struct netstat)*num_if);
  netstat_new=(struct netstat *)malloc(sizeof(struct netstat)*num_if);
  netstat_last=(struct netstat *)malloc(sizeof(struct netstat)*num_if);
  ifaceslist =(struct str_liptr *)malloc(sizeof(struct str_liptr)*num_if*16+1);

  net_fd=open("/proc/net/dev",O_RDONLY);
  /* skip legend */
  getstr(net_fd);
  getstr(net_fd);

  i=0;
  while(buf=getstr(net_fd)){
    /* sscanf(buf," %s:",netstat_new[i].if_name); */
    start=0;
    while(buf[start]==' ' && buf[start]!='\0')
      ++start;
    j=start;
    while(buf[j]!=':' && buf[j]!='\0')
      ++j;
    buf[j]='\0';
    strcpy(netstat_new[i].if_name,buf+start);
    /*if(debug){fprintf(stderr,"ifname%d: '%s'\n",i,netstat_new[i].if_name);}*/
    strcpy(netstat_cur[i].if_name,netstat_new[i].if_name);
    strcpy(netstat_last[i].if_name,netstat_new[i].if_name);
    netstat_last[i].rcv_bytes=
      netstat_last[i].rcv_pckts=
      netstat_last[i].rcv_errs=
      netstat_last[i].rcv_drop=
      netstat_last[i].rcv_fifo=
      netstat_last[i].rcv_frame=
      netstat_last[i].rcv_compr=
      netstat_last[i].rcv_multi=
      netstat_last[i].sent_bytes=
      netstat_last[i].sent_pckts=
      netstat_last[i].sent_errs=
      netstat_last[i].sent_drop=
      netstat_last[i].sent_fifo=
      netstat_last[i].sent_colls=
      netstat_last[i].sent_carr=
      netstat_last[i].sent_compr=0;

    ifaceslist[i*16+ 0].str=malloc(12+IF_NAME_LEN);
    ifaceslist[i*16+ 1].str=malloc(12+IF_NAME_LEN);
    ifaceslist[i*16+ 2].str=malloc(12+IF_NAME_LEN);
    ifaceslist[i*16+ 3].str=malloc(12+IF_NAME_LEN);
    ifaceslist[i*16+ 4].str=malloc(12+IF_NAME_LEN);
    ifaceslist[i*16+ 5].str=malloc(12+IF_NAME_LEN);
    ifaceslist[i*16+ 6].str=malloc(12+IF_NAME_LEN);
    ifaceslist[i*16+ 7].str=malloc(12+IF_NAME_LEN);
    ifaceslist[i*16+ 8].str=malloc(12+IF_NAME_LEN);
    ifaceslist[i*16+ 9].str=malloc(12+IF_NAME_LEN);
    ifaceslist[i*16+10].str=malloc(12+IF_NAME_LEN);
    ifaceslist[i*16+11].str=malloc(12+IF_NAME_LEN);
    ifaceslist[i*16+12].str=malloc(12+IF_NAME_LEN);
    ifaceslist[i*16+13].str=malloc(12+IF_NAME_LEN);
    ifaceslist[i*16+14].str=malloc(12+IF_NAME_LEN);
    ifaceslist[i*16+15].str=malloc(12+IF_NAME_LEN);
    sprintf(ifaceslist[i*16+ 0].str,"%s_rcv_bytes",netstat_new[i].if_name);
    sprintf(ifaceslist[i*16+ 1].str,"%s_rcv_pckts",netstat_new[i].if_name);
    sprintf(ifaceslist[i*16+ 2].str,"%s_rcv_errs",netstat_new[i].if_name);
    sprintf(ifaceslist[i*16+ 3].str,"%s_rcv_drop",netstat_new[i].if_name);
    sprintf(ifaceslist[i*16+ 4].str,"%s_rcv_fifo",netstat_new[i].if_name);
    sprintf(ifaceslist[i*16+ 5].str,"%s_rcv_frame",netstat_new[i].if_name);
    sprintf(ifaceslist[i*16+ 6].str,"%s_rcv_compr",netstat_new[i].if_name);
    sprintf(ifaceslist[i*16+ 7].str,"%s_rcv_multi",netstat_new[i].if_name);
    sprintf(ifaceslist[i*16+ 8].str,"%s_sent_bytes",netstat_new[i].if_name);
    sprintf(ifaceslist[i*16+ 9].str,"%s_sent_pckts",netstat_new[i].if_name);
    sprintf(ifaceslist[i*16+10].str,"%s_sent_errs",netstat_new[i].if_name);
    sprintf(ifaceslist[i*16+11].str,"%s_sent_drop",netstat_new[i].if_name);
    sprintf(ifaceslist[i*16+12].str,"%s_sent_fifo",netstat_new[i].if_name);
    sprintf(ifaceslist[i*16+13].str,"%s_sent_colls",netstat_new[i].if_name);
    sprintf(ifaceslist[i*16+14].str,"%s_sent_carr",netstat_new[i].if_name);
    sprintf(ifaceslist[i*16+15].str,"%s_sent_compr",netstat_new[i].if_name);
    ifaceslist[i*16+ 0].ptr=&netstat_cur[i].rcv_bytes;
    ifaceslist[i*16+ 1].ptr=&netstat_cur[i].rcv_pckts;
    ifaceslist[i*16+ 2].ptr=&netstat_cur[i].rcv_errs;
    ifaceslist[i*16+ 3].ptr=&netstat_cur[i].rcv_drop;
    ifaceslist[i*16+ 4].ptr=&netstat_cur[i].rcv_fifo;
    ifaceslist[i*16+ 5].ptr=&netstat_cur[i].rcv_frame;
    ifaceslist[i*16+ 6].ptr=&netstat_cur[i].rcv_compr;
    ifaceslist[i*16+ 7].ptr=&netstat_cur[i].rcv_multi;
    ifaceslist[i*16+ 8].ptr=&netstat_cur[i].sent_bytes;
    ifaceslist[i*16+ 9].ptr=&netstat_cur[i].sent_pckts;
    ifaceslist[i*16+10].ptr=&netstat_cur[i].sent_errs;
    ifaceslist[i*16+11].ptr=&netstat_cur[i].sent_drop;
    ifaceslist[i*16+12].ptr=&netstat_cur[i].sent_fifo;
    ifaceslist[i*16+13].ptr=&netstat_cur[i].sent_colls;
    ifaceslist[i*16+14].ptr=&netstat_cur[i].sent_carr;
    ifaceslist[i*16+15].ptr=&netstat_cur[i].sent_compr;

    ++i;
  }
  ifaceslist[num_if*16].str=NULL;
  ifaceslist[num_if*16].ptr=NULL;

  /* just initialize "last" values */
  get_info_stat();
  get_info_net();
  get_info_nfs();
  get_info_nfsd();
  get_info_loadavg();
  get_info_mem();

  /* make i/o nonblocked */
  flags=fcntl(0,F_GETFD,0);
  fcntl(0,F_SETFD,flags|O_NONBLOCK);

  flags=fcntl(1,F_GETFD,0);
  fcntl(1,F_SETFD,flags|O_NONBLOCK);


  /* first read = init */
  for(;;){
    if(read(0,&p,1)<1){
      usleep(100);
      continue;
    }
    if(p=='\n')
      break;
  }

  print("cpu_user\ncpu_nice\ncpu_system\ncpu_idle\npgpin\npgpout\nswpin\nswpout\ncontext\nmem_total\nmem_free\n"
        "mem_shared\nmem_buff\nmem_cached\nmem_swp\nmem_swp_free\nnfs_count\nnfs_retrans\nnfs_auth\n"
        "nfs2_null\nnfs2_getattr\nnfs2_setattr\nnfs2_root\nnfs2_lookup\nnfs2_readlink\nnfs2_read\n"
        "nfs2_writechache\nnfs2_write\nnfs2_create\nnfs2_remove\nnfs2_rename\nnfs2_link\nnfs2_symlink\n"
        "nfs2_mkdir\nnfs2_rmdir\nnfs2_readdir\nnfs2_statf\nnfs3_null\nnfs3_getattr\nnfs3_setattr\n"
        "nfs3_lookup\nnfs3_access\nnfs3_readlink\nnfs3_read\nnfs3_write\nnfs3_create\nnfs3_mkdir\n"
        "nfs3_symlink\nnfs3_mknod\nnfs3_remove\nnfs3_rmdir\nnfs3_rename\nnfs3_link\nnfs3_readdir\n"
        "nfs3_readdirplus\nnfs3_fsstat\nnfs3_fsinfo\nnfs3_pathconf\nnfs3_commit\nnfsd_calls\n"
        "nfsd_badcalls\nnfsd_badauth\nnfsd_badclnt\nnfsd_xdrcall\nnfsd2_null\nnfsd2_getattr\n"
        "nfsd2_setattr\nnfsd2_root\nnfsd2_lookup\nnfsd2_readlink\nnfsd2_read\nnfsd2_writechache\n"
        "nfsd2_write\nnfsd2_create\nnfsd2_remove\nnfsd2_rename\nnfsd2_link\nnfsd2_symlink\nnfsd2_mkdir\n"
        "nfsd2_rmdir\nnfsd2_readdir\nnfsd2_statfs\nnfsd3_null\nnfsd3_getattr\nnfsd3_setattr\nnfsd3_lookup\n"
        "nfsd3_access\nnfsd3_readlink\nnfsd3_read\nnfsd3_write\nnfsd3_create\nnfsd3_mkdir\nnfsd3_symlink\n"
        "nfsd3_mknod\nnfsd3_remove\nnfsd3_rmdir\nnfsd3_rename\nnfsd3_link\nnfsd3_readdir\nnfsd3_readdirplus\n"
        "nfsd3_fsstat\nnfsd3_fsinfo\nnfsd3_pathconf\nnfsd3_commit\nnfsd3_cache\nnfsd3_miss\nnfsd3_nocache\n"
        "nfsd3_io_read\nnfsd3_io_write\nnfsd3_netcount\nnfsd3_thread_sec\nnfsd3_thread_10\n"
        "nfsd3_thread_20\nnfsd3_thread_30\nnfsd3_thread_40\nnfsd3_thread_50\nnfsd3_thread_60\n"
        "nfsd3_thread_70\nnfsd3_thread_80\nnfsd3_thread_90\nnfsd3_thread_100\n");

  if(stat_long)
    print("cpu_sys_real\ncpu_iowait\ncpu_irq\ncpu_softirq\ncpu_steal\n");

  for(i=0;i<num_cpu;++i){
    print("cpu_user%d\ncpu_nice%d\ncpu_system%d\ncpu_idle%d\n",i,i,i,i);
    if(stat_long)
      print("cpu_sys_real%d\ncpu_iowait%d\ncpu_irq%d\ncpu_softirq%d\ncpu_steal%d\n",i,i,i,i,i);
  }

  for(i=0;i<num_if;++i){
    print("%s_rcv_bytes\n%s_rcv_pckts\n%s_rcv_errs\n%s_rcv_drop\n%s_rcv_fifo\n%s_rcv_frame\n"
          "%s_rcv_compr\n%s_rcv_multi\n%s_sent_bytes\n%s_sent_pckts\n%s_sent_errs\n"
          "%s_sent_drop\n%s_sent_fifo\n%s_sent_colls\n%s_sent_carr\n%s_sent_compr\n",
          netstat_new[i].if_name,netstat_new[i].if_name,
          netstat_new[i].if_name,netstat_new[i].if_name,
          netstat_new[i].if_name,netstat_new[i].if_name,
          netstat_new[i].if_name,netstat_new[i].if_name,
          netstat_new[i].if_name,netstat_new[i].if_name,
          netstat_new[i].if_name,netstat_new[i].if_name,
          netstat_new[i].if_name,netstat_new[i].if_name,
          netstat_new[i].if_name,netstat_new[i].if_name);
  }
  print("forks\n\n");

  /* main loop */
  for(;;){

    i=0;
    if(debug){
      fprintf(stderr,"---------------\n");
    }

    /* reset readed chars!!! */
    p=p2='\0';

    /* read request */
    for(;;){

      /* one-by-one byte */
      while(read(0,&p,1)>0){
        if(p == '\n'){
          if(p2 == '\n'){
            request[i]='\0';
            break;
          }
        }
        request[i++]=p;
        p2=p;
        if(i>MAX_REQ_LINE)
          exit(2);
      }
      if((p == '\n') && (p2 == '\n')){
        break;
      }
      usleep(100);
    }

    /* update data */
    get_info_stat();
    get_info_net();
    get_info_nfs();
    get_info_nfsd();
    get_info_loadavg();
    get_info_mem();

    if(debug){
      fprintf(stderr,"SS: '%s'\n",request);
    }
    /* process request ! */
    req=0;
    for(i=0;request[i]!='\0';){
      while(request[i]!='\n')
        ++i;
      request[i]='\0';
      ++i;
      if(debug){
        fprintf(stderr,"SS<- '%s'\n",request+req);
      }
      process_line(request+req);
      req=i;
    }
    print("\n");
  }

  return 0;
}
